###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_AS_Client_Type_DataStorage."""
import unittest

dummyResults = {}


class DummyReturn:
    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self.dummyMethod

    def dummyMethod(self, *args, **kwargs):
        return dummyResults[self.__class__.__name__]


class dBaseAccountingType(DummyReturn):
    pass


################################################################################


class DataStorage_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        # We need the proper software, and then we overwrite it.
        import LHCbDIRAC.AccountingSystem.Client.Types.DataStorage as moduleTested

        moduleTested.BaseAccountingType = dBaseAccountingType
        moduleTested.DataStorage.__bases__ = (dBaseAccountingType,)

        self.accountingType = moduleTested.DataStorage

    def tearDown(self):
        """TearDown."""
        del self.accountingType


class DataStorage_Success(DataStorage_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        global dummyResults
        dummyResults["DataStorage"] = None

        accountingType = self.accountingType()
        self.assertEqual("DataStorage", accountingType.__class__.__name__)


################################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DataStorage_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DataStorage_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: RAWIntegrityHandler

.. module: RAWIntegrityHandler

:synopsis: DISET interface to the RAWIntegrityDB.
"""
# imports

# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler

# from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.DB.RAWIntegrityDB import RAWIntegrityDB

# global instance of the RAWIntegrityDB
gRAWIntegrityDB = False


def initializeRAWIntegrityHandler(_serviceInfo):
    """handler initialisation."""

    global gRAWIntegrityDB
    gRAWIntegrityDB = RAWIntegrityDB()

    # Create DB tables if needed
    return gRAWIntegrityDB._checkTable()


class RAWIntegrityHandler(RequestHandler):
    """
    .. class:: RAWIntegrityHandler
    """

    types_addFile = [
        (str,),
        (str,),
        (int,),
        (str,),
        (str,),
        (str,),
    ]

    @staticmethod
    def export_addFile(lfn, pfn, size, se, guid, checksum):
        """Add a file to the RAW integrity DB."""
        try:
            gLogger.info(f"RAWIntegrityHandler.addFile: Attempting to add {lfn} to the gRAWIntegrityDB.")
            res = gRAWIntegrityDB.addFile(lfn, pfn, size, se, guid, checksum)
            return res
        except Exception as x:
            errStr = "RAWIntegrityHandler.addFile: Exception while adding file to gRAWIntegrityDB."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    types_getGlobalStatistics = []

    @staticmethod
    def export_getGlobalStatistics():
        """Get global file statistics."""
        try:
            gLogger.info("Attempting to get global statistics.")
            res = gRAWIntegrityDB.getGlobalStatistics()
            if not res["OK"]:
                gLogger.error("getGlobalStatistics: Failed to get global statistics", res["Message"])
            else:
                gLogger.info("getGlobalStatistics: Obtained global statistics")
            return res
        except Exception as x:
            errStr = "getGlobalStatistics: Exception while getting global statistics."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    types_getFileSelections = []

    @staticmethod
    def export_getFileSelections():
        """Get the possible selections available."""
        try:
            gLogger.info("Attempting to get selections.")
            res = gRAWIntegrityDB.getFileSelections()
            if not res["OK"]:
                gLogger.error("getFileSelections: Failed to get file selections", res["Message"])
            else:
                gLogger.info("getFileSelections: Obtained file selections")
            return res
        except Exception as x:
            errStr = "getFileSelections: Exception while getting file selections."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    types_getFilesSummaryWeb = [dict, list, (int,), (int,)]

    @staticmethod
    def export_getFilesSummaryWeb(selectDict, sortList, startItem, maxItems):
        """Get the file information according to conventions."""
        resultDict = {}
        startDate = selectDict.get("FromDate", None)
        if startDate:
            del selectDict["FromDate"]
        endDate = selectDict.get("ToDate", None)
        if endDate:
            del selectDict["ToDate"]
        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = sortList[0][0] + ":" + sortList[0][1]
        else:
            orderAttribute = None
        res = gRAWIntegrityDB.selectFiles(selectDict, orderAttribute=orderAttribute, newer=startDate, older=endDate)
        if not res["OK"]:
            return S_ERROR("Failed to select jobs: " + res["Message"])
        # Get the files and the counters correctly
        fileList = res["Value"]
        nFiles = len(fileList)
        resultDict["TotalRecords"] = nFiles
        if nFiles == 0:
            return S_OK(resultDict)
        iniFile = startItem
        lastFile = iniFile + maxItems
        if iniFile >= nFiles:
            return S_ERROR("Item number out of range")
        if lastFile > nFiles:
            lastFile = nFiles
        summaryFileList = fileList[iniFile:lastFile]
        # Prepare the standard format
        resultDict["ParameterNames"] = [
            "lfn",
            "pfn",
            "size",
            "storageelement",
            "guid",
            "checksum",
            "startTime",
            "endTime",
            "status",
        ]
        records = []
        statusCountDict = {}
        for lfn, pfn, size, se, guid, checksum, submit, complete, status in fileList:
            if status not in statusCountDict:
                statusCountDict[status] = 0
            statusCountDict[status] += 1
        for fileTuple in summaryFileList:
            lfn, pfn, size, se, guid, checksum, submit, complete, status = fileTuple
            startTime = str(submit)
            endTime = str(complete)
            records.append((lfn, pfn, size, se, guid, checksum, startTime, endTime, status))
        resultDict["Records"] = records
        resultDict["Extras"] = statusCountDict
        return S_OK(resultDict)

    types_getStatistics = [(str,), dict]

    @staticmethod
    def export_getStatistics(attribute, selectDict):
        """Get job statistics distribution per attribute value with a given
        selection."""
        startDate = selectDict.get("FromDate", None)
        if startDate:
            del selectDict["FromDate"]
        # For backward compatibility
        if startDate is None:
            startDate = selectDict.get("LastUpdate", None)
            if startDate:
                del selectDict["LastUpdate"]
        endDate = selectDict.get("ToDate", None)
        if endDate:
            del selectDict["ToDate"]

        result = gRAWIntegrityDB.getCounters("Files", [attribute], selectDict, newer=startDate, older=endDate)

        resultDict = {}
        if result["OK"]:
            for cDict, count in result["Value"]:
                resultDict[cDict[attribute]] = count

        return S_OK(resultDict)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import sys

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.parseCommandLine()

    import DIRAC

    transIDs = []
    if len(sys.argv) < 2:
        print("Usage: dirac-transformation-clean transID [transID] [transID]")
        DIRAC.exit(1)
    else:
        try:
            transIDs = [int(arg) for arg in sys.argv[1:]]
        except Exception:
            print("Invalid list of transformations")
            DIRAC.exit(1)

    from LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent

    agent = TransformationCleaningAgent(
        "Transformation/TransformationCleaningAgent",
        "Transformation/TransformationCleaningAgent",
        "dirac-transformation-clean",
    )
    agent.initialize()

    for transID in transIDs:
        agent.cleanTransformation(transID)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to remove input data files for given workflow.

Initially written for use after merged outputs have been successfully
uploaded to an SE.
"""
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class RemoveInputData(ModuleBase):
    #############################################################################

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("RemoveInputData")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        # List all parameters here
        self.inputDataList = []

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""

        super()._resolveInputVariables()

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution function."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            if not self._checkWFAndStepStatus():
                return S_OK()

            if not self._enableModule():
                return S_OK()

            self._resolveInputVariables()

            # Try to remove the file list with failover if necessary
            failover = []
            self.log.info(f'Attempting dm.removeFile("{self.inputDataList}")')
            result = self.dataManager.removeFile(self.inputDataList)
            self.log.verbose(result)
            if not result["OK"]:
                self.log.error(
                    'Could not remove files with message:\n"%s"\n\
        Will set removal requests just in case.'
                    % (result["Message"])
                )
                failover = self.inputDataList
            try:
                if result["Value"]["Failed"]:
                    failureDict = result["Value"]["Failed"]
                    if failureDict:
                        self.log.info(
                            f'Not all files were successfully removed, see "LFN : reason" below\n{failureDict}'
                        )
                    failover = list(failureDict)
            except KeyError:
                self.log.error(f"Setting files for removal request to be the input data: {self.inputDataList}")
                failover = self.inputDataList

            for lfn in failover:
                self.__setFileRemovalRequest(lfn)

            return S_OK("Input Data Removed")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in RemoveInputData execute module", lException=e)
            self.setApplicationStatus(e)
            return S_ERROR(str(e))

        finally:
            super().finalize()

    #############################################################################

    def __setFileRemovalRequest(self, lfn):
        """Sets a removal request for a file including all replicas."""
        self.log.info(f"Setting file removal request for {lfn}")
        removeFile = Operation()
        removeFile.Type = "RemoveFile"
        rmFile = File()
        rmFile.LFN = lfn
        removeFile.addFile(rmFile)
        self.request.addOperation(removeFile)



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody


class DataChallengeReplicationBody(BaseBody):
    """
    This body is to be used for the Data challenge, and will be
    changed regularly

    * The file is originally in CERN-DAQ-EXPORT
    * We replicate it to CERN-DC-RAW (waiting for the file to be archived)
    * We then replicate to the chosen Tier 1 disk (DataChallenge shares).
    * We replicate it from the Tier1-disk to the Tier1-Tape
    * Finally, we remove the replica on the Tier1-Disk
    """

    # This is needed to know how to serialize the object
    _attrToSerialize = []

    @staticmethod
    def _createOperation(operationType, lfns, targetSE, sourceSE=None):
        """
        Generate an Operation object of the given Type, with the
        specified targetSE and sourceSE. Associate to it File objects
        from the lfns list

        :param str operationType: Type of the operation (RemoveReplica, ReplicateAndRegister, etc)
        :param list lfns: list of LFNs on which to perform the operation
        :param str targetSE: SE name(s) targeted by the ops
        :param str sourceSE: SE name of the source
        """
        newOp = Operation()
        newOp.Type = operationType
        newOp.TargetSE = targetSE
        if sourceSE:
            newOp.SourceSE = sourceSE
        for lfn in lfns:
            opFile = File()
            opFile.LFN = lfn
            newOp.addFile(opFile)
        return newOp

    def taskToRequest(self, taskID, task, transID):
        """
        Create the request object from the task Dict


        """

        if isinstance(task["InputData"], list):
            lfns = task["InputData"]
        elif isinstance(task["InputData"], str):
            lfns = task["InputData"].split(";")
        else:
            raise ValueError(f"InputData must be a list or a string, not {type(task['InputData'])}")

        req = Request()

        # # Replicate first to CERN-DC-RAW
        # repCernRAW = self._createOperation("ReplicateAndRegister", lfns, "CERN-DC-RAW")
        # req.addOperation(repCernRAW)

        # Now look at the target SEs
        # We expect one "Site-BUFFER" SE and one "Site-RAW" SE
        targetSEs = task["TargetSE"].split(",")

        # Order them to have the buffer first
        bufferSE, rawSE = sorted(targetSEs, key=lambda se: "-BUFFER" in se, reverse=True)

        # First replicate to disk
        repBuffer = self._createOperation("ReplicateAndRegister", lfns, bufferSE)
        req.addOperation(repBuffer)

        # Then replicate to tape from disk
        repTape = self._createOperation("ReplicateAndRegister", lfns, rawSE, sourceSE=bufferSE)
        req.addOperation(repTape)

        # Finally, remove the replicas from Tier1-Disk
        daqExportRemoval = self._createOperation("RemoveReplica", lfns, bufferSE)
        req.addOperation(daqExportRemoval)

        return req



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
   LHCbDIRAC - LHCb extension of DIRAC

   References:
    DIRAC: https://github.com/DIRACGrid/DIRAC
    LHCbDIRAC: https://gitlab.cern.ch/lhcb-dirac/LHCbDIRAC

   The distributed data production and analysis system of LHCb.
"""
import os

from pkg_resources import get_distribution, DistributionNotFound


rootPath = os.path.dirname(os.path.realpath(__path__[0]))

# Define Version
try:
    __version__ = get_distribution(__name__).version
    version = __version__
except DistributionNotFound:
    # package is not installed
    version = "Unknown"


def extension_metadata():
    return {
        "primary_extension": True,
        "priority": 100,
        "setups": {
            "LHCb-Production": "dips://lhcb-conf-dirac.cern.ch:9135/Configuration/Server",
            "LHCb-Certification": "https://lhcb-cert-dirac.cern.ch:8443/Configuration/Server",
        },
        "default_setup": "LHCb-Production",
    }



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the storage usage summary for the given directories."""
from DIRAC.Core.Base.Script import Script

seSvcClassDict = {}
infoStringLength = 1
unit = None
dmScript = None


def seSvcClass(se):
    if se not in seSvcClassDict:
        from DIRAC.Resources.Storage.StorageElement import StorageElement

        try:
            if se.endswith("HIST"):
                seSvcClassDict[se] = "Hist"
            else:
                status = StorageElement(se).getStatus()
                if status["OK"]:
                    status = status["Value"]
                    seSvcClassDict[se] = "Tape" if status["TapeSE"] else "Disk" if status["DiskSE"] else "Unknown"
                else:
                    seSvcClassDict[se] = "Unknown"
        except BaseException:
            seSvcClassDict[se] = "LFN"
    return seSvcClassDict[se]


def orderSEs(listSEs):
    """Order SEs: LFNs, then tape Ses then disk SEs The input list can be a list
    a dict, a tuple or a set."""
    orderedSEs = ["LFN"] if "LFN" in sorted(listSEs) else []
    orderedSEs += sorted([se for se in listSEs if se not in orderedSEs and se.endswith("-HIST")])
    orderedSEs += sorted([se for se in listSEs if se not in orderedSEs and (seSvcClass(se) == "Tape")])
    orderedSEs += sorted([se for se in listSEs if se not in orderedSEs])
    return orderedSEs


def printSEUsage(totalUsage, grandTotal, scaleFactor):
    """Nice printout of SE usage."""
    dashes = "-" * 48
    print(dashes)
    print(f"{'DIRAC SE'.ljust(20)} {(f'Size ({unit})').ljust(20)} {'Files'.ljust(20)}")
    form = "%.1f"
    orderedSEs = orderSEs(totalUsage)
    for se in orderedSEs:
        if totalUsage[se]["Size"] / scaleFactor < 1.0:
            form = "%.3f"
            break
    svcClass = ""
    sumFiles = 0
    sumSize = 0.0
    for se in orderedSEs:
        newSvcClass = seSvcClass(se)
        if newSvcClass != svcClass:
            if svcClass:
                print(
                    "%s %s %s"
                    % (("Total (%s)" % svcClass).ljust(20), (form % (sumSize)).ljust(20), str(sumFiles).ljust(20))
                )
                sumFiles = 0
                sumSize = 0.0
            print(dashes)
        svcClass = newSvcClass
        usageDict = totalUsage[se]
        files = usageDict["Files"]
        size = usageDict["Size"] / scaleFactor
        sumFiles += files
        sumSize += size
        print(f"{se.ljust(20)} {(form % size).ljust(20)} {str(files).ljust(20)}")
    if grandTotal:
        size = grandTotal["Size"] / scaleFactor
        print(f"{'Total (disk)'.ljust(20)} {(form % size).ljust(20)} {str(grandTotal['Files']).ljust(20)}")
    print(dashes)


def printBigTable(siteList, bigTable):
    """Print out THE big table of usage."""
    siteList.sort()
    for site in ("ARCHIVE", "LFN"):
        if site in siteList:
            siteList.remove(site)
            siteList.insert(0, site)
    just = [20, 30, 15]
    for cond in bigTable:
        just[0] = max(just[0], len(cond) + 1)
        for processingPass in sorted(bigTable[cond]):
            just[1] = max(just[1], len(processingPass) + 1)
    prStr = "Conditions".ljust(just[0]) + "ProcessingPass".ljust(just[1])
    for site in siteList:
        prStr += site.ljust(just[2])
    print(prStr)
    grandTotal = {}
    for cond in sorted(bigTable):
        print(cond.ljust(just[0]))
        for processingPass in sorted(bigTable[cond]):
            prStr = "".ljust(just[0]) + processingPass.ljust(just[1])
            bigTableUsage = bigTable[cond][processingPass][1]
            for site in siteList:
                if site in bigTableUsage:
                    grandTotal[site] = grandTotal.setdefault(site, 0) + bigTableUsage[site]
                    prStr += f"{bigTableUsage[site]:.3f}".ljust(just[2])
                elif site == "Total":
                    prStr += "0".ljust(just[2])
                else:
                    prStr += "-".ljust(just[2])
            print(prStr)
    prStr = "\n" + "".ljust(just[0]) + "Grand-Total".ljust(just[1])
    for site in siteList:
        if site in grandTotal:
            prStr += f"{grandTotal[site]:.3f}".ljust(just[2])
        else:
            prStr += "-".ljust(just[2])
    print(prStr)


def writeInfo(str):
    global infoStringLength
    import sys

    if len(str) < 1 or str[0] != "\n":
        sys.stdout.write(" " * infoStringLength + "\r")
    sys.stdout.write(str + "\r")
    sys.stdout.flush()
    infoStringLength = len(str) + 1


def browseBK(bkQuery, ses, scaleFactor):
    # Cannot be imported at the top because CS init
    from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE

    bkPath = bkQuery.getPath()
    if not bkQuery.getConfiguration():
        print(f"The Configuration should be specified in the --BKQuery option: {bkPath}")
        return None
    conditions = bkQuery.getBKConditions()
    if not conditions:
        print(f"No Conditions found for this Configuration {bkPath}")
        return None
    requestedEventTypes = bkQuery.getEventTypeList()
    requestedFileTypes = bkQuery.getFileTypeList()
    requestedPP = bkQuery.getProcessingPass()
    requestedConditions = bkQuery.getConditions()
    bigTable = {}
    siteList = []
    for cond in conditions:
        strCond = f"Browsing conditions {cond}"
        writeInfo(strCond)
        bkQuery.setConditions(cond)
        eventTypesForPP = bkQuery.getBKProcessingPasses()
        # print eventTypesForPP
        for processingPass in [pp for pp in eventTypesForPP if eventTypesForPP[pp]]:
            if requestedEventTypes:
                eventTypes = [t for t in requestedEventTypes if t in eventTypesForPP[processingPass]]
                if not eventTypes:
                    continue
            else:
                eventTypes = eventTypesForPP[processingPass]
            strPP = f" - ProcessingPass {processingPass}"
            writeInfo(strCond + strPP)
            bkQuery.setProcessingPass(processingPass)
            totalUsage = {}
            grandTotal = {}
            allProds = []
            # print eventTypes
            for eventType in eventTypes:
                strEvtType = f" - EventType {str(eventType)}"
                writeInfo(strCond + strPP + strEvtType)
                bkQuery.setEventType(eventType)
                fileTypes = bkQuery.getBKFileTypes()
                bkQuery.setFileType(fileTypes)
                # print fileTypes
                if not fileTypes or None in fileTypes:
                    continue
                prods = sorted(bkQuery.getBKProductions())
                # print prods
                strProds = f" - FileTypes {str(fileTypes)} - Prods {str(prods)}"
                writeInfo(strCond + strPP + strEvtType + strProds)
                info = bkQuery.getNumberOfLFNs()
                nbFiles = info["NumberOfLFNs"]
                size = info["LFNSize"]
                totalUsage["LFN"]["Size"] = totalUsage.setdefault("LFN", {}).setdefault("Size", 0) + size
                totalUsage["LFN"]["Files"] = totalUsage["LFN"].setdefault("Files", 0) + nbFiles
                if "DST" not in fileTypes:
                    fileTypes.append("DST")
                if "MDST" not in fileTypes:
                    fileTypes.append("MDST")
                if prods:
                    allProds += prods
                    for prodID in prods:
                        totalUsage, grandTotal = getStorageSummary(totalUsage, grandTotal, "", fileTypes, prodID, ses)
                bkQuery.setFileType(requestedFileTypes)
                bkQuery.setEventType(requestedEventTypes)
            writeInfo("")
            if allProds:
                allProds.sort()
                print(cond, processingPass, allProds)
                printSEUsage(totalUsage, grandTotal, scaleFactor)
                processingPass = processingPass.replace("/Real Data", "")
                bigTable.setdefault(cond, {})[processingPass] = [allProds, {}]
                bigTableUsage = bigTable[cond][processingPass][1]
                for se in totalUsage:
                    if se.endswith("-ARCHIVE"):
                        site = "ARCHIVE"
                    elif se == "LFN":
                        site = "LFN"
                    else:
                        res = getSitesForSE(se, gridName="LCG")
                        if res["OK"] and len(res["Value"]) > 0:
                            site = res["Value"][0]
                        else:
                            continue
                    if site not in siteList:
                        siteList.append(site)
                    bigTableUsage[site] = bigTableUsage.setdefault(site, 0) + (totalUsage[se]["Size"] / scaleFactor)
            bkQuery.setProcessingPass(requestedPP)
        bkQuery.setConditions(requestedConditions)
    import datetime

    print("\n", bkQuery.getPath(), str(datetime.datetime.today()).split()[0])
    printBigTable(siteList, bigTable)


def getStorageSummary(totalUsage, grandTotal, dirName, fileTypes, prodID, ses):
    from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

    if not totalUsage:
        totalUsage = {}
    if not grandTotal:
        grandTotal = {"Files": 0, "Size": 0}
    if not isinstance(fileTypes, type([])):
        fileTypes = [fileTypes]
    for fileType in fileTypes:
        res = StorageUsageClient().getStorageSummary(dirName, fileType, prodID, ses)
        if res["OK"]:
            for se in res["Value"]:
                totalUsage.setdefault(se, {"Files": 0, "Size": 0})
                totalUsage[se]["Files"] += res["Value"][se]["Files"]
                totalUsage[se]["Size"] += res["Value"][se]["Size"]
                if seSvcClass(se) == "Disk" and not se.endswith("-HIST"):
                    grandTotal["Files"] += res["Value"][se]["Files"]
                    grandTotal["Size"] += res["Value"][se]["Size"]
    return totalUsage, grandTotal


def execute(unit, minimum, depth):
    import DIRAC
    from DIRAC import gConfig, gLogger
    from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

    # gLogger.setLevel( 'FATAL' )
    lcg = False
    full = False
    topDirectories = False
    bkBrowse = False
    users = None
    summary = False
    rank = False
    ses = []
    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "u" or switch[0].lower() == "unit":
            unit = switch[1]
        elif switch[0].lower() == "l" or switch[0].lower() == "lcg":
            lcg = True
        elif switch[0].lower() == "full":
            full = True
        elif switch[0] == "TopDirectories":
            if not topDirectories:
                topDirectories = depth
        elif switch[0] == "Depth":
            topDirectories = int(switch[1])
        elif switch[0] == "BrowseBK":
            bkBrowse = True
        elif switch[0] == "Users":
            users = switch[1].lower()
            if users == "all":
                summary = True
                rank = True
            else:
                users = sorted(users.split(","))
        elif switch[0] == "Summary":
            summary = True
        elif switch[0] == "Minimum":
            minimum = float(switch[1])
        elif switch[0] == "UnknownSE":
            ses = switch[1].split(",")

    scaleDict = {
        "MB": 1000 * 1000.0,
        "GB": 1000 * 1000 * 1000.0,
        "TB": 1000 * 1000 * 1000 * 1000.0,
        "PB": 1000 * 1000 * 1000 * 1000 * 1000.0,
    }
    if unit not in scaleDict:
        Script.showHelp()
    scaleFactor = scaleDict[unit]

    ses += dmScript.getOption("SEs", [])
    sites = dmScript.getOption("Sites", [])
    for site in sites:
        res = gConfig.getOptionsDict(f"/Resources/Sites/LCG/{site}")
        if not res["OK"]:
            print(f"Site {site} not known")
            Script.showHelp()
        ses.extend(res["Value"]["SE"].replace(" ", "").split(","))

    # Create a bkQuery looking at all files
    if bkBrowse:
        bkQuery = dmScript.getBKQuery(visible="All")
        browseBK(bkQuery, ses, scaleFactor)
        DIRAC.exit(0)

    rpc = StorageUsageClient()
    dirs = dmScript.getOption("Directory", [])
    if users:
        if users != "all":
            dirs += [f"/lhcb/user/{user[0]}/{user}" for user in users]
            minimum = 0
        else:
            res = rpc.getStorageDirectories("/lhcb/user", None, None, None)
            if not res["OK"]:
                print("Error getting directories in /lhcb/user:", res["Message"])
                DIRAC.exit(2)
            dirs += sorted({"/".join(d.split("/")[0:5]) for d in res["Value"]})
            users = [user for user in [d.split("/")[-1] for d in dirs] if user]
    prods = None
    # print bkQuery
    fileTypes = dmScript.getOption("FileType", [])
    if not dirs:
        dirs = ["/lhcb/"]
        bkQuery = dmScript.getBKQuery()
        if bkQuery:
            bkQuery = dmScript.getBKQuery(visible="All")
            bkFileTypes = bkQuery.getFileTypeList()
            if bkFileTypes:
                fileTypes = bkFileTypes
            print("BK query:", bkQuery)
            if fileTypes == ["RAW"]:
                # For RAW data, get the list of directories...
                dirs = bkQuery.getDirs()
            else:
                prods = sorted(bkQuery.getBKProductions())
                if not prods:
                    print(f"No productions found for bkQuery {str(bkQuery)}")
                    DIRAC.exit(0)
                # As storageSummary deals with directories and not real file types,
                #    add DST in order to cope with old naming convention
                if fileTypes and "FULL.DST" not in fileTypes and "DST" not in fileTypes:
                    fileTypes.append("DST")
                print(f"Looking for {len(prods)} productions:", prods)
        elif fileTypes and fileTypes[0]:
            print("FileTypes:", fileTypes)

    if not prods:
        prods = [""]
    if not fileTypes:
        fileTypes = [""]
    prString = "Storage usage for "
    if sites:
        prString += f"Sites {str(sites)} "
    elif ses:
        prString += f"SEs {str(ses)} "
    if prods[0] != "":
        prString += f"productions {str(prods)} "
    if fileTypes[0] not in ("", None):
        prString += f"file types {str(fileTypes)} "
    if dirs[0] != "":
        if fileTypes == ["RAW"]:
            prString += f"in {len(dirs)} directories"
        else:
            prString += f"directories {str(dirs)}"
    if not users:
        if prString == "Storage usage for ":
            prString += "all SEs"
        gLogger.notice(prString)
    if full or topDirectories:
        dirData = {}
        for prodID in sorted(prods):
            for fileType in fileTypes:
                for dirName in dirs:
                    res = rpc.getStorageDirectoryData(dirName, fileType, prodID, ses)
                    if not res["OK"]:
                        print("Failed to get directories", res["Message"])
                        DIRAC.exit(2)
                    dirData.update(res["Value"])
        if full:
            for resDir in sorted(dirData):
                gLogger.notice(resDir + " -> ", str(dirData[resDir]))
        if topDirectories:
            gLogger.notice("Depth-4 directories:")
            topDirData = {}
            for resDir, usage in dirData.items():
                topDir = "/".join(resDir.split("/")[: topDirectories + 1]) + "/"
                topDirData.setdefault(topDir, {"Files": 0, "Size": 0})
                topDirData[topDir]["Files"] += usage["Files"]
                topDirData[topDir]["Size"] += usage["Size"]
            for resDir in sorted(topDirData):
                gLogger.notice(resDir + " -> ", str(topDirData[resDir]))

    totalUsage = None
    grandTotal = None
    usersUsage = {}
    for prodID in prods:
        for fileType in fileTypes:
            for dirName in dirs:
                if users:
                    user = dirName.split("/")[-1]
                    userRegistry = gConfig.getOptions(f"/Registry/Users/{user}")
                    if not userRegistry["OK"]:
                        quota = 0
                    else:
                        quota = gConfig.getValue(f"/Registry/Users/{user}/Quota", 0)
                        if not quota:
                            quota = gConfig.getValue("/Registry/DefaultStorageQuota", 0)
                    quota *= scaleDict["GB"] / scaleFactor
                    totalUsage, grandTotal = getStorageSummary(0, 0, dirName, fileType, prodID, ses)
                    spaceUsed = grandTotal["Size"] / scaleFactor
                    if summary:
                        usersUsage[user] = (spaceUsed, quota)
                    else:
                        print(
                            "Storage usage for user %s (quota: %.1f %s)%s"
                            % (
                                user,
                                quota,
                                unit,
                                (
                                    " <== User no longer registered"
                                    if not quota
                                    else (" <== Over quota" if spaceUsed > quota else "")
                                ),
                            )
                        )
                        printSEUsage(totalUsage, grandTotal, scaleFactor)
                else:
                    totalUsage, grandTotal = getStorageSummary(totalUsage, grandTotal, dirName, fileType, prodID, ses)

    if lcg:
        # Cannot be imported at the top because CS init
        from DIRAC.Resources.Storage.StorageElement import StorageElement

        tapeTotalFiles = 0
        diskTotalFiles = 0
        tapeTotalSize = 0
        diskTotalSize = 0
        for se in sorted(totalUsage):
            storageElement = StorageElement(se)
            files = totalUsage[se]["Files"]
            size = totalUsage[se]["Size"]
            seStatus = storageElement.getStatus()
            if seStatus["OK"]:
                seStatus = seStatus["Value"]
                if seStatus["TapeSE"]:
                    tapeTotalFiles += files
                    tapeTotalSize += size
                if seStatus["DiskSE"]:
                    diskTotalFiles += files
                    diskTotalSize += size
        print(f"{'Storage Type'.ljust(20)} {(f'Size ({unit})').ljust(20)} {'Files'.ljust(20)}")
        print("-" * 50)
        print(
            "%s %s %s"
            % ("T1D*".ljust(20), ("%.1f" % (tapeTotalSize / scaleFactor)).ljust(20), str(tapeTotalFiles).ljust(20))
        )
        print(
            "%s %s %s"
            % ("T*D1".ljust(20), ("%.1f" % (diskTotalSize / scaleFactor)).ljust(20), str(diskTotalFiles).ljust(20))
        )
        DIRAC.exit(0)

    if not users:
        printSEUsage(totalUsage, grandTotal, scaleFactor)
    elif summary:
        if rank:
            users = sorted(usersUsage, key=lambda x: float(usersUsage[x][0]), reverse=True)
        for user in users:
            spaceUsed, quota = usersUsage[user]
            if spaceUsed > minimum:
                print(
                    "Storage usage for user %8s: %6.3f %s (quota: %4.1f %s)%s"
                    % (
                        user,
                        spaceUsed,
                        unit,
                        quota,
                        unit,
                        (
                            " <== User no longer registered"
                            if not quota
                            else (" <== Over quota" if spaceUsed > quota else "")
                        ),
                    )
                )

    DIRAC.exit(0)


@Script()
def main():
    global unit
    global dmScript

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerNamespaceSwitches()
    dmScript.registerSiteSwitches()

    unit = "TB"
    minimum = 0.1
    depth = 4
    Script.registerSwitch("u:", "Unit=", f"   Unit to use [{unit}] (MB,GB,TB,PB)")
    Script.registerSwitch("l", "LCG", "  Group results by tape and disk")
    Script.registerSwitch("", "Full", "  Output the directories matching selection")
    Script.registerSwitch("", "TopDirectories", "  Same as --Full but group by depth-%d directories" % depth)
    Script.registerSwitch("", "Depth=", "   Specify the depth for top directories (default: %d)" % depth)
    Script.registerSwitch("", "BrowseBK", "   Loop overall paths matching the BK query")
    Script.registerSwitch("", "Users=", "   Get storage usage for a (list of) user(s)")
    Script.registerSwitch("", "Summary", "  Only print a summary for users")
    Script.registerSwitch(
        "", "Minimum=", f"   Don't print usage for users below that usage (same units, default {minimum:.1f}"
    )
    Script.registerSwitch("", "UnknownSE=", "   Force using a non-existing SE name")

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)

    from DIRAC import gConfig, gLogger

    execute(unit, minimum, depth)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" pytest for BookkeepingClient
"""
# pylint: disable=protected-access,missing-docstring,invalid-name

import pytest
from unittest.mock import MagicMock

# sut
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

bk = BookkeepingClient()

# mocks
mockRPC = MagicMock()
mockTC = MagicMock()
mockPR = MagicMock()

# test data
Step1 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_1", "gauss-conddb_1", None, 1, "Y")
Step2 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_2", "gauss-conddb_2", None, 2, "Y")
Step3 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "fromPreviousStep", "fromPreviousStep", None, 3, "Y")
Step3Corr_1 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_1", "gauss-conddb_1", None, 3, "Y")
Step3Corr_2 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_2", "gauss-conddb_2", None, 3, "Y")
Step4 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "fromPreviousStep", "fromPreviousStep", None, 4, "Y")
Step4Corr_1 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_1", "gauss-conddb_1", None, 4, "Y")
Step4Corr_2 = ("gauss", "Gauss", "v1r1", "/some/gauss/option/files", "gauss-dddb_2", "gauss-conddb_2", None, 4, "Y")


# Actual tests


@pytest.mark.parametrize(
    "mockTCResponse, mockPRResponse, expected, expectedRes",
    [
        ({"OK": False, "Message": "bof"}, {}, False, None),
        ({"OK": True, "Value": {"Status": "Archived"}}, {}, False, None),
        ({"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}}, {"OK": False, "Message": "bif"}, False, None),
        ({"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}}, {"OK": True, "Value": [1]}, True, []),
        ({"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}}, {"OK": True, "Value": [2, 1]}, True, [2]),
        ({"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}}, {"OK": True, "Value": [3, 2, 1]}, True, [2, 3]),
        ({"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}}, {"OK": True, "Value": [3, 1, 2]}, True, [3]),
        (
            {"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}},
            {"OK": True, "Value": [4, 3, 1, 2]},
            True,
            [3, 4],
        ),
        (
            {"OK": True, "Value": {"RequestID": 1, "Status": "Cleaned"}},
            {"OK": True, "Value": [4, 3, 1, 2]},
            False,
            None,
        ),
    ],
)
def test__getPreviousProductions(mocker, mockTCResponse, mockPRResponse, expected, expectedRes):
    mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.TransformationClient", return_value=mockTC)
    mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.ProductionRequestClient", return_value=mockPR)

    mockTC.getTransformation.return_value = mockTCResponse
    mockPR.getProductionList.return_value = mockPRResponse

    res = bk._getPreviousProductions(1)
    assert res["OK"] is expected
    if res["OK"]:
        assert res["Value"] == expectedRes


@pytest.mark.parametrize(
    "steps, expected",
    [
        ([], []),
        ([Step1], [Step1]),
        ([Step1, Step2], [Step1, Step2]),
        ([Step1, Step2, Step3], [Step1, Step2, Step3Corr_2]),
        ([Step1, Step2, Step3, Step4], [Step1, Step2, Step3Corr_2, Step4Corr_2]),
        ([Step1, Step3, Step2, Step4], [Step1, Step3Corr_1, Step2, Step4Corr_2]),
        ([Step1, Step4, Step3, Step2], [Step1, Step4Corr_1, Step3Corr_1, Step2]),
        ([Step3, Step4, Step2, Step1], [Step3, Step4, Step2, Step1]),
        ([Step3], [Step3]),
    ],
)
def test__resolveProductionSteps(steps, expected):
    res = bk._resolveProductionSteps(steps)
    assert res == expected


@pytest.mark.parametrize(
    "mockRPCResponse, mockTCResponse, mockPRResponse, expected, expectedRes",
    [
        ({"OK": False, "Message": "bif"}, {}, {}, False, {}),
        ({"OK": True, "Value": []}, {}, {}, False, {}),
        ({"OK": True, "Value": [Step1]}, {}, {}, True, [Step1]),
        ({"OK": True, "Value": [Step1, Step2]}, {}, {}, True, [Step1, Step2]),
        ({"OK": True, "Value": [Step1, Step3]}, {}, {}, True, [Step1, Step3Corr_1]),
        ({"OK": True, "Value": [Step1, Step3, Step4]}, {}, {}, True, [Step1, Step3Corr_1, Step4Corr_1]),
        (
            {"OK": True, "Value": [Step3]},
            {"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}},
            {"OK": True, "Value": [2, 1]},
            True,
            [Step3],
        ),
        (
            {"OK": True, "Value": [Step3, Step4]},
            {"OK": True, "Value": {"RequestID": 1, "Status": "Archived"}},
            {"OK": True, "Value": [3, 2, 1]},
            True,
            [Step3, Step4],
        ),
    ],
)
def test_getSteps(mocker, mockRPCResponse, mockTCResponse, mockPRResponse, expected, expectedRes):
    mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.Client._getRPC", return_value=mockRPC)
    mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.TransformationClient", return_value=mockTC)
    mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.ProductionRequestClient", return_value=mockPR)

    mockRPC.getSteps.return_value = mockRPCResponse
    mockTC.getTransformation.return_value = mockTCResponse
    mockPR.getProductionList.return_value = mockPRResponse

    res = bk.getSteps(1)
    assert res["OK"] is expected
    if res["OK"]:
        assert res["Value"] == expectedRes



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = ["start_productions", "check_validation"]

import traceback
import json

from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import parse_obj
from LHCbDIRAC.ProductionManagementSystem.scripts.dirac_production_request_submit import submitProductionRequests
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import configure_input, runs_to_input_query

from .integrations import Validation, Request, RepoIssue


def get_request_state(request_id):
    return returnValueOrRaise(ProductionRequestClient().getProductionRequest([request_id]))[int(request_id)][
        "RequestState"
    ]


def update_request_state(request_id, state):
    returnValueOrRaise(ProductionRequestClient().updateProductionRequest(int(request_id), {"RequestState": state}))
    print(f"Request {request_id} set to {state}")


def start_productions(logbook, issue):
    if "state::run-validation" in issue.labels:
        validation = True
    elif "state::ready" in issue.labels:
        if not issue.metadata.validations:
            raise NotImplementedError("No validation found")
        for request_id, request_info in issue.metadata.validations.items():
            if request_info.running:
                raise NotImplementedError()
            for result in returnValueOrRaise(
                TransformationClient().getTransformations(
                    {"TransformationID": request_info.transform_ids}, columns=["TransformationID", "Status"]
                )
            ):
                if result["Status"] not in {"Cleaning", "Cleaned"}:
                    raise NotImplementedError(f"Request {request_id} is not cleaned")
            current_state = get_request_state(request_id)
            if current_state == "Active":
                update_request_state(request_id, "Done")
            elif current_state not in {"Done", "Rejected"}:
                raise NotImplementedError(f"Request {request_id} is in state {current_state}")
        validation = False
    else:
        raise NotImplementedError(issue.labels)

    productionRequest = parse_obj(issue.request_yaml)
    match productionRequest.type:
        case "Sprucing":
            launch_func = launch_sprucing
        case _:
            raise NotImplementedError(productionRequest.type)
    if validation:
        productionRequest.name = f"Validation - {productionRequest.name}"
    productionIDs = submitProductionRequests([productionRequest], dryRun=False, createFiletypes=True)

    request_id = productionIDs[0][1]

    print("Submitted request", request_id)
    transform_ids, n_files_expected = launch_func(
        request_id,
        validation=validation,
        start_run=issue.run_yaml["start_run"],
        end_run=issue.run_yaml["end_run"],
    )

    try:
        if validation:
            if request_id in issue.metadata.validations:
                raise NotImplementedError()
            issue.metadata.validations[request_id] = Validation(
                transform_ids=transform_ids,
                n_files_expected=n_files_expected,
            )
        else:
            issue.labels.pop(issue.labels.index("state::ready"))
            issue.labels.append("state::running-concurrent" if issue.run_yaml["concurrent"] else "state::running")
            issue.metadata.requests[request_id] = Request(transform_ids=transform_ids)
        issue.update_metadata()
    except Exception:
        print("Failed to update metadata in issue", issue.metadata)
        traceback.print_exc()

    subject = productionRequest.name
    if validation:
        body = f"Validation production was submitted with request ID: {request_id}"
    else:
        body = f"Production was submitted with request ID: {request_id}"
    if validation:
        subject = f"{subject} - Validation"
    attachments = {
        f"{request_id}-request.yaml": issue.request_yaml_blob,
        f"{request_id}-extra.yaml": issue.run_yaml_blob,
    }
    try:
        logbook.create_post(issue.url, transform_ids, subject, body, attachments, ["Sprucing", "Production"])
    except Exception:
        print("Failed to post to logbook")
        traceback.print_exc()

    try:
        update_request_state(request_id, "Active")
    except Exception:
        print("Failed to set request to active")
        traceback.print_exc()


def start_removal_transformation(logbook, issue: RepoIssue, request_id: int):
    if "state::running" not in issue.labels:
        raise NotImplementedError("Only submitting removals for issues in state::running")

    request_metadata = issue.metadata.requests[request_id]
    sprucing_tid, merging_tid = request_metadata.transform_ids

    tc = TransformationClient()
    bkquery = returnValueOrRaise(tc.getBookkeepingQuery(sprucing_tid))
    bkquery["ProcessingPass"] = "/" + bkquery["ProcessingPass"]
    extra_param = returnValueOrRaise(tc.getAdditionalParameters(sprucing_tid))
    removal_trans = Transformation()
    removal_trans.setType("Removal")

    trans_name = "/".join(
        [
            "Removal-",
            bkquery["ConfigName"],
            bkquery["ConfigVersion"],
            bkquery["DataTakingConditions"],
            bkquery["ProcessingPass"],
            str(bkquery["EventType"]),
            f"{bkquery['FileType']}-issue-{issue.issue.iid}",
        ]
    )
    removal_trans.setBkQuery(bkquery)
    removal_trans.setTransformationName(trans_name)
    removal_trans.setTransformationGroup("RemoveReplicasWhenProcessed")
    removal_trans.setPlugin("RemoveReplicasWhenProcessed")
    removal_trans.setBody("removal;RemoveReplica")
    removal_trans.setSEParam("FromSEs", "['Tier1-Buffer']")
    long_name = f"RemoveReplicasWhenProcessed for issue {issue.issue.iid}"
    removal_trans.setDescription(long_name[:255])
    removal_trans.setLongDescription(long_name[:255])
    removal_trans.setAdditionalParam("ProcessingPasses", extra_param["groupDescription"].strip("/"))
    # removal_trans.setStatus("Active")
    removal_trans.setAgentType("Automatic")
    returnValueOrRaise(removal_trans.addTransformation())
    removal_id = returnValueOrRaise(removal_trans.getTransformationID())
    print(f"Submitted removal transforation {request_id=} {removal_id=}", request_id)

    try:

        issue.metadata.requests[request_id].removal = removal_id
        issue.update_metadata()
    except Exception:
        print("Failed to update metadata in issue", issue.metadata)
        traceback.print_exc()

    productionRequest = parse_obj(issue.request_yaml)
    subject = f"Removal for {productionRequest.name}"
    body = f"Removal transformation was submitted: {removal_id}\n\n"

    body += json.dumps(removal_trans.paramValues, indent=3)

    try:
        logbook.create_post(issue.url, [removal_id], subject, body, {}, ["Sprucing", "Data Management", "Production"])
    except Exception:
        print("Failed to post to logbook")
        traceback.print_exc()


def check_validation(issue, prod_id):
    request_metadata = issue.metadata.validations[prod_id]

    if not request_metadata.running:
        raise NotImplementedError()

    result = returnValueOrRaise(
        TransformationClient().getTransformations(
            {"TransformationID": request_metadata.transform_ids}, columns=["TransformationID", "Status"]
        )
    )
    statuses = {x["Status"] for x in result}
    if statuses.issubset({"Cleaning", "Cleaned"}):
        request_metadata.running = False
        request_metadata.cleaned = True
        issue.update_metadata()
        update_request_state(prod_id, "Done")
        print("Request is cleaned, validation will need to be resubmitted")
        return

    query = {"TransformationID": request_metadata.transform_ids}
    results = {}
    stats = returnValueOrRaise(TransformationClient().getTransformations(query, columns=["TransformationID", "Status"]))
    for stat in stats:
        results[stat.pop("TransformationID")] = stat | {"FileStatus": {}}
    file_stats = returnValueOrRaise(
        TransformationClient().getCounters("TransformationFiles", ["TransformationID", "Status"], query)
    )
    for meta, count in file_stats:
        results[meta["TransformationID"]]["FileStatus"][meta["Status"]] = count
    is_idle = all(
        x["Status"] == "Idle" and "Assigned" not in x["FileStatus"] and sum(x["FileStatus"].values()) > 0
        for x in results.values()
    )
    is_complete = all(set(x["FileStatus"]) == {"Processed"} for x in results.values())
    print(f"Request {prod_id}: {request_metadata.n_files_expected=} {is_idle=} {is_complete=}: {results}")
    if not is_idle:
        print("Validation is still running")
    elif is_complete:
        if request_metadata.n_files_expected != results[min(results)]["FileStatus"]["Processed"]:
            raise NotImplementedError()
        message = (
            """The validation is complete. Please check the results and update the ~"state::ready" when appropriate."""
        )
        issue.discussions.create({"body": message})
        issue.labels.pop(issue.labels.index("state::run-validation"))
        issue.labels.append("state::check-validation")
        request_metadata.running = False
        issue.update_metadata()
        print("Validation is complete")
    else:
        print("Something went wrong, please check the status of the transformations!")


def launch_sprucing(
    request_id: int,
    *,
    validation: bool,
    start_run: int | None = None,
    end_run: int | None = None,
    append_name: str = "1",
    ancestor_depth: int = 0,
):
    prods = returnValueOrRaise(ProductionRequestClient().getProductionRequest([request_id]))
    data = prods[request_id]
    n_files_expected = None

    pr = ProductionRequest()
    pr.requestID = str(request_id)
    pr.appendName = append_name
    pr.visibility = "Yes"
    kwargs = {} if validation else dict(startRun=start_run, endRun=end_run)
    configure_input(pr, data, **kwargs)
    pr.prodGroup = f"{pr.processingPass}/{json.loads(data['ProDetail'])['pDsc']}"
    pr.outConfigName = "validation" if validation else pr.configName
    if validation:
        bkQueryDict = pr._getBKKQuery().copy()
        if "RunNumbers" in bkQueryDict:
            bkQueryDict["RunNumbers"] = bkQueryDict["RunNumbers"].split(";;;")
        if "DataQualityFlag" in bkQueryDict:
            bkQueryDict["DataQualityFlag"] = bkQueryDict["DataQualityFlag"].split(";;;")
        if "ExtendedDQOK" in bkQueryDict:
            bkQueryDict["ExtendedDQOK"] = bkQueryDict["ExtendedDQOK"].split(";;;")
        all_runs = sorted(
            run
            for run in returnValueOrRaise(BookkeepingClient().getListOfRuns(bkQueryDict))
            if start_run <= run <= end_run
        )
        # Run validations with the first run to ensure it will have been distributed
        # Originally we tried to use first + random + last run however when sites are
        # full the last run might not be distributed and the validation will be stuck.
        runs = [all_runs.pop(0)]
        runs = sorted(set(runs))
        print("Runs to be validated: ", runs)
        n_files_expected = 0
        for run in runs:
            result = returnValueOrRaise(BookkeepingClient().getFilesWithMetadata(bkQueryDict | {"RunNumbers": run}))
            if result["TotalRecords"] == 0:
                raise NotImplementedError(f"Run {run} has no files")
            print(f"Run {run} has {result['TotalRecords']} files")
            n_files_expected += result["TotalRecords"]
        pr.runsList = ",".join(runs_to_input_query(runs))
    configure_steps(pr, data, ancestor_depth)
    transform_ids = returnValueOrRaise(pr.buildAndLaunchRequest())
    return transform_ids, n_files_expected


def configure_steps(pr, data, ancestor_depth):
    pd = json.loads(data["ProDetail"])
    steps = {int(k[1 : -len("Step")]): v for k, v in pd.items() if k.endswith("Step")}
    pr.stepsList = [int(v) for k, v in sorted(steps.items())]
    pr.prodsTypeList = ["Sprucing", "Merge"]
    pr.stepsInProds = [list(range(1, len(pr.stepsList))), [len(pr.stepsList)]]
    pr.targets = ["", ""]
    pr.outputSEs = ["Tier1-Buffer", "Tier1-DST"]
    pr.specialOutputSEs = [{}, {}]
    pr.removeInputsFlags = [False, True]
    pr.priorities = [2, 8]
    pr.inputs = [[], []]
    pr.inputDataPolicies = ["download", "download"]
    pr.multicore = ["True", "True"]
    pr.outputModes = ["Run", "Run"]
    pr.targets = ["ALL", "ALL"]
    pr.events = ["-1", "-1"]
    pr.ancestorDepths = [ancestor_depth, 0]
    pr.compressionLvl = ["LOW", "HIGH"]
    pr.plugins = ["Sprucing", "ByRunFileTypeSizeWithFlush"]
    pr.groupSizes = [1, 5]
    pr.cpus = [1_000_000, 300_000]
    pr.outputVisFlag = [{"1": "N"}, {"2": "Y"}]
    pr.specialOutputVisFlag = [{"1": {}}, {"2": {}}]



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve files of a given type for a production."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... ProdID Type",
                "Arguments:",
                "  ProdID:   Production ID (integer)",
                "  Type:     File Type (For example: ALL, DST, SIM, DIGI, RDST, MDF)",
            ]
        )
    )
    Script.parseCommandLine()
    args = Script.getPositionalArgs()

    if not len(args) == 2:
        Script.showHelp(exitCode=1)

    try:
        prodID = int(args[0])
    except BaseException:
        Script.showHelp(exitCode=1)
    filetype = args[1]

    from DIRAC import gLogger, exit as DIRACexit
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    client = BookkeepingClient()
    res = client.getProductionFiles(prodID, filetype)
    if not res["OK"]:
        gLogger.error("ERROR: Failed to retrieve production files", res["Message"])
        DIRACexit(1)

    if not res["Value"]:
        gLogger.notice(f"No files found for production {prodID} with type {filetype}")
    else:
        gLogger.notice(
            "%s %s %s %s %s"
            % ("FileName".ljust(100), "Size".ljust(10), "GUID".ljust(40), "Replica".ljust(8), "Visible".ljust(8))
        )
        for lfn in sorted(res["Value"]):
            size = res["Value"][lfn]["FileSize"]
            guid = res["Value"][lfn]["GUID"]
            hasReplica = res["Value"][lfn]["GotReplica"]
            visible = res["Value"][lfn]["Visible"]
            gLogger.notice(
                "%s %s %s %s %s"
                % (lfn.ljust(100), str(size).ljust(10), guid.ljust(40), str(hasReplica).ljust(8), str(visible).ljust(8))
            )


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It stores the replica."""
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.Replica import Replica
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.ReplicaParam import ReplicaParam
from DIRAC import gLogger


class ReplicaReader:
    @staticmethod
    def readReplica(doc, filename):
        """reads the replica information."""
        gLogger.debug("Reading Replica from" + str(filename))
        replica = Replica()
        replica.name = filename  # full path

        replicaElements = doc.getElementsByTagName("Replica")

        for node in replicaElements:
            param = ReplicaParam()

            outputfile = node.getAttributeNode("File")
            if outputfile:
                param.file = outputfile.value
            else:
                gLogger.warn("Missing the <file> tag in replica xml file")

            name = node.getAttributeNode("Name")
            if name:
                param.name = name.value
            else:
                gLogger.warn("Missing the <name> tag in replica xml file")

            location = node.getAttributeNode("Location")
            if location:
                param.location = location.value
            else:
                gLogger.warn("Missing the <location> tag in replica xml file")

            se = node.getAttributeNode("SE")
            if se:
                param.se = se.value
            else:
                gLogger.warn("Missing the <SE> tag in replica xml file")

            action = node.getAttributeNode("Action")
            if action:
                param.action = action.value
            else:
                gLogger.warn("Missing the <Action> tag in replica xml file")

            replica.addParam(param)
            gLogger.info("Replica Reading finished successfully")

        return replica



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""List all BK paths matching a wildcard path ('...' is the wildcard character,
or '*' but enclose with quotes)"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import Script, ProgressBar
    from DIRAC import gLogger

    Script.registerSwitch("B:", "BKQuery=", "   Bookkeeping query path")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    bkPaths = []
    for opt, val in Script.getUnprocessedSwitches():
        if opt == "BKQuery":
            bkPaths = val.split(",")
    if not bkPaths:
        gLogger.error("No BK path provided...")
        Script.showHelp(exitCode=1)

    from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import getProcessingPasses, BKQuery

    for i, bkPath in enumerate(bkPaths):
        if i:
            gLogger.notice("=========================")
        bkQuery = BKQuery(bkPath.replace("Real Data", "RealData"))
        progressBar = ProgressBar(1, title=f"Getting processing passes for BK path {bkPath}")
        processingPasses = getProcessingPasses(bkQuery)
        progressBar.endLoop()
        if processingPasses:
            gLogger.notice(
                "\n".join([""] + [procPass.replace("Real Data", "RealData") for procPass in sorted(processingPasses)])
            )
        else:
            gLogger.notice("No processing passes matching the BK path")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set of functions used by the DMS checking scripts."""
import os
from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import (
    removeFiles,
    removeReplicas,
    registerBK2FC,
    printDMResult,
)


def __removeFile(lfns):
    """Use the ScriptExecutors removeFile method."""
    if isinstance(lfns, str):
        lfns = [lfns]
    removeFiles(lfns)


def __removeReplica(lfnDict):
    """Use the ScriptExecutors removeReplicas method."""
    seLFNs = {}
    for lfn in lfnDict:
        for se in lfnDict[lfn]:
            seLFNs.setdefault(se, []).append(lfn)
    for se, lfns in seLFNs.items():
        removeReplicas(lfns, [se])
    return seLFNs


def __replaceReplica(dm, seLFNs):
    """Re-replicate replicas that had just been removed because they were bad It
    uses the DataManager instance from the ConsistencyCheck."""
    if seLFNs:
        gLogger.notice("Now replicating bad replicas...")
        success = {}
        failed = {}
        for se, lfns in seLFNs.items():
            for lfn in lfns:
                res = dm.replicateAndRegister(lfn, se)
                if res["OK"]:
                    success.update(res["Value"]["Successful"])
                    failed.update(res["Value"]["Failed"])
                else:
                    failed[lfn] = res["Message"]

        failures = 0
        errors = {}
        for lfn, reason in failed.items():
            reason = str(reason)
            errors[reason] = errors.setdefault(reason, 0) + 1
            failures += 1
        gLogger.notice("\t%d success, %d failures%s" % (len(success), failures, ":" if failures else ""))
        for reason in errors:
            gLogger.notice("\tError %s : %d files" % (reason, errors[reason]))


def _dumpErrorAndFiles(title, lfnList, maxFiles, dumpStr, fileName, fp):
    """Print out errors and a (restricted) list of LFNs dump the whole list in a
    file."""
    gLogger.error(title)
    if not gLogger.info("\n".join(sorted(lfnList))):
        if len(lfnList) > maxFiles:
            gLogger.error("First %d files:" % maxFiles)
        gLogger.error("\n".join(sorted(lfnList)[0:maxFiles]))
    if fileName is None:
        return None
    if fp is None:
        fp = open(fileName, "w")
    dumpStr1 = dumpStr.strip().replace("\n", "")
    if not dumpStr.endswith(" "):
        dumpStr += " "
    if not dumpStr.startswith("\n"):
        dumpStr = "\n" + dumpStr
    fp.write(dumpStr)
    fp.write(dumpStr.join([""] + sorted(lfnList)))
    gLogger.notice(f"Whole list available with 'grep {dumpStr1} {fileName}'")
    return fp


def _getUniqueFileName(name):
    """Get a file name that doesn't exist given a prefix."""
    suffix = ""
    nb = 0
    while True:
        fileName = name + f"{suffix}.txt"
        if not os.path.exists(fileName):
            break
        nb += 1
        suffix = "-%d" % nb
    return fileName


def doCheckFC2SE(cc, bkCheck=True, fixIt=False, replace=False, maxFiles=None, fixOption="FixIt"):
    """Method actually calling for the the check using ConsistencyChecks module
    It prints out results and calls corrective actions if required."""
    cc.checkFC2SE(bkCheck)

    if maxFiles is None:
        maxFiles = 20
    fileName = _getUniqueFileName("CheckFC2SE")
    fp = None

    if cc.inSEbutNotInFC:
        gLogger.notice(">>>>")
        gLogger.notice("Some files found in SE but not in FC")
        if fixIt:
            gLogger.notice("Going to register files from BK to FC")
            regResult = {"Successful": [], "Failed": {}}
        for lfn, seList in cc.inSEbutNotInFC.items():
            inBK = lfn in cc.existLFNsBKRepNo or lfn in cc.existLFNsBKRepYes
            gLogger.notice(f"\t{lfn} ({'in BK' if inBK else 'not in BK'}): {','.join(seList)}")
            if fixIt:
                if inBK:
                    res = registerBK2FC([lfn], seList)
                    if res["OK"]:
                        regResult["Successful"] += res["Value"]["Successful"]
                        regResult["Failed"].update(res["Value"]["Failed"])
                    else:
                        regResult["Failed"][lfn] = res["Message"]
                else:
                    regResult["Failed"][lfn] = "Not in BK"

        if fixIt:
            # We can remove these files from the BK-No list as we just set it back
            for lfn in regResult["Successful"]:
                cc.existLFNsBKRepNo.pop(lfn, None)
            printDMResult(S_OK(regResult))
        else:
            gLogger.notice("Use --FixIt to register files to the FC if they are in BK")
        gLogger.notice("<<<<")
    elif cc.seList:
        gLogger.notice("None of the files absent in FC were found at specified SEs")

    if cc.notRegisteredAtSE:
        gLogger.notice(">>>>")
        gLogger.notice("Some files found in SE but not registered in FC")
        if fixIt:
            gLogger.notice("Going to register files in FC")
            registrationProtocol = DMSHelpers().getRegistrationProtocols()
            dm = DataManager()
            replicaTuples = []
            regRepResult = {"Successful": {}, "Failed": {}}
            for lfn, seList in cc.notRegisteredAtSE.items():
                for se in seList:
                    res = StorageElement(se).getURL(lfn, protocol=registrationProtocol)
                    if res["OK"]:
                        pfn = res["Value"]["Successful"][lfn]
                        replicaTuples.append((lfn, pfn, se))
            res = dm.registerReplica(replicaTuples)
            if res["OK"]:
                regRepResult["Successful"].update(res["Value"]["Successful"])
                regRepResult["Failed"].update(res["Value"]["Failed"])
            else:
                regRepResult["Failed"][lfn] = res["Message"]
            printDMResult(S_OK(regRepResult))
        else:
            gLogger.notice("Use --FixIt to register replicas in the FC")
        gLogger.notice("<<<<")

    if cc.existLFNsBKRepNo:
        gLogger.notice(">>>>")
        affectedRuns = {str(run) for run in cc.existLFNsBKRepNo.values() if run}
        title = "%d files are in the FC (or SE) but have replica = NO in BK:\nAffected runs: %s" % (
            len(cc.existLFNsBKRepNo),
            ",".join(sorted(affectedRuns)) if affectedRuns else "None",
        )
        fp = _dumpErrorAndFiles(title, cc.existLFNsBKRepNo, maxFiles, "InFCButBKNo", fileName, fp)
        if fixIt:
            gLogger.notice("Going to fix them, setting the replica flag")
            res = cc.bkClient.addFiles(list(cc.existLFNsBKRepNo))
            if res["OK"]:
                gLogger.notice("\tSuccessfully added replica flag")
            else:
                gLogger.error("Failed to set the replica flag", res["Message"])
        else:
            gLogger.notice(f"Use --{fixOption} to fix it (set the replica flag)")
        gLogger.notice("<<<<")

    elif bkCheck:
        gLogger.notice("No files in FC with replica = NO in BK -> OK!")

    if cc.existLFNsNotInBK:
        gLogger.notice(">>>>")

        title = f"{len(cc.existLFNsNotInBK)} files are in the FC but are NOT in BK"
        fp = _dumpErrorAndFiles(title, cc.existLFNsNotInBK, maxFiles, "InFCButNotInBK ", fileName, fp)
        if fixIt:
            gLogger.notice("Going to fix them, by removing from the FC and storage")
            __removeFile(cc.existLFNsNotInBK)
        else:
            gLogger.notice(f"Use --{fixOption} to fix it (remove from FC and storage)")
        gLogger.notice("<<<<")

    else:
        gLogger.notice("No files in FC not in BK -> OK!")

    seOK = True
    if cc.existLFNsNoSE:
        gLogger.notice(">>>>")

        seOK = False
        nOfExistLFNsNoSE = len(cc.existLFNsNoSE)
        title = f"{nOfExistLFNsNoSE} file{'' if nOfExistLFNsNoSE == 1 else 's'}"
        title += f" {'is' if nOfExistLFNsNoSE == 1 else 'are'} in the BK and FC"
        title += f" but {'is' if nOfExistLFNsNoSE == 1 else 'are'} missing at some SEs"
        fixStr = "removing them from catalogs" if not replace else "re-replicating them"
        fp = _dumpErrorAndFiles(title, cc.existLFNsNoSE, maxFiles, "InFCButNotInSE ", fileName, fp)
        if fixIt:
            gLogger.notice("Going to fix, " + fixStr)
            removeLfns = []
            replicasToRemove = {}
            for lfn, ses in cc.existLFNsNoSE.items():
                if ses == "All":
                    removeLfns.append(lfn)
                else:
                    replicasToRemove.setdefault(lfn, []).extend(ses)
            if removeLfns:
                __removeFile(removeLfns)
            if replicasToRemove:
                seLFNs = __removeReplica(replicasToRemove)
                if replace:
                    __replaceReplica(cc.dataManager, seLFNs)
        else:
            if not replace:
                fixStr += "; use --Replace if you want to re-replicate them"
            gLogger.notice(f"Use --{fixOption} to fix it ({fixStr})")
        gLogger.notice("<<<<")

    else:
        gLogger.notice("No missing replicas at sites -> OK!")

    if cc.existLFNsBadFiles or cc.existLFNsNotExisting:
        gLogger.notice(">>>>")

        seOK = False
        if cc.existLFNsBadFiles:
            title = f"{len(cc.existLFNsBadFiles)} files have a bad checksum"
            fp = _dumpErrorAndFiles(title, cc.existLFNsBadFiles, maxFiles, "BadFiles", fileName, fp)
        if cc.existLFNsNotExisting:
            title = f"{len(cc.existLFNsNotExisting)} files don't exist at any SE"
            fp = _dumpErrorAndFiles(title, cc.existLFNsNotExisting, maxFiles, "LostFiles", fileName, fp)
        toRemove = sorted(set(cc.existLFNsBadFiles) | set(cc.existLFNsNotExisting))
        if fixIt:
            gLogger.notice("Going to fix them, removing files from catalogs and storage")
            __removeFile(toRemove)
        else:
            gLogger.notice(f"Use --{fixOption} to fix it (remove files from SE and catalogs)")
        gLogger.notice("<<<<")

    if cc.existLFNsBadReplicas:
        gLogger.notice(">>>>")

        seOK = False
        title = f"{len(cc.existLFNsBadReplicas)} replicas have a bad checksum"
        badChecksum = [f"{lfn} @ {','.join(sorted(se))}" for lfn, se in cc.existLFNsBadReplicas.items()]
        fp = _dumpErrorAndFiles(title, badChecksum, maxFiles, "BadChecksum", fileName, fp)
        fixStr = "remove replicas from SE and catalogs" if not replace else "re-replicating them"
        if fixIt:
            gLogger.notice("Going to fix, " + fixStr)
            seLFNs = __removeReplica(cc.existLFNsBadReplicas)
            if replace:
                __replaceReplica(cc.dataManager, seLFNs)
        else:
            if not replace:
                fixStr += "; use --Replace if you want to re-replicate them"
            gLogger.notice(f"Use --{fixOption} to fix it ({fixStr})")
        gLogger.notice("<<<<")

    if not cc.existLFNsBadFiles and not cc.existLFNsBadReplicas:
        gLogger.notice("No replicas have a bad checksum -> OK!")
    if seOK:
        gLogger.notice("All files exist and have a correct checksum -> OK!")


def doCheckFC2BK(cc, fixFC=False, fixBK=False, listAffectedRuns=False, checkFC2SE=True):
    """Method actually calling for the the check using ConsistencyChecks module
    It prints out results and calls corrective actions if required."""
    cc.checkFC2BK()

    maxFiles = 10
    suffix = ""
    nb = 0
    baseName = "CheckFC2BK" + (f"-{cc.prod}" if cc.prod else "")
    while True:
        fileName = baseName + f"{suffix}.txt"
        if not os.path.exists(fileName):
            break
        nb += 1
        suffix = "-%d" % nb
    fp = None
    if cc.existLFNsBKRepNo:
        gLogger.notice(">>>>")

        affectedRuns = list({str(run) for run in cc.existLFNsBKRepNo.values()})
        nOfExistLFNsBKRepNo = len(cc.existLFNsBKRepNo)
        errMsg = f"{nOfExistLFNsBKRepNo} file{'' if nOfExistLFNsBKRepNo == 1 else 's'} "
        errMsg += f"{'is' if nOfExistLFNsBKRepNo == 1 else 'are'} in the FC but have replica = NO in BK"
        gLogger.error(errMsg)
        if checkFC2SE:
            from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

            ccAux = ConsistencyChecks()
            gLogger.notice(f"====== Now checking {nOfExistLFNsBKRepNo} files from FC to SE ======")
            ccAux.lfns = list(cc.existLFNsBKRepNo)
            doCheckFC2SE(ccAux, bkCheck=False, fixIt=fixFC, fixOption="FixFC")
            cc.existLFNsBKRepNo = set(cc.existLFNsBKRepNo)
            cc.existLFNsBKRepNo -= set(ccAux.existLFNsNoSE)
            cc.existLFNsBKRepNo -= set(ccAux.existLFNsNotExisting)
            cc.existLFNsBKRepNo -= set(ccAux.existLFNsBadFiles)
        cc.existLFNsBKRepNo = sorted(cc.existLFNsBKRepNo)
        if cc.existLFNsBKRepNo:
            msg = f"====== Completed, {nOfExistLFNsBKRepNo} file{'s' if nOfExistLFNsBKRepNo == 1 else 's'} "
            msg += f"{'is' if nOfExistLFNsBKRepNo == 1 else 'are'} in the FC and SE but have replica = NO in BK ======"
            gLogger.notice(msg)
            if fp is None:
                fp = open(fileName, "w")
            fp.write("\nInFCButBKNo ".join([""] + sorted(cc.existLFNsBKRepNo)))
            res = cc.bkClient.getFileMetadata(cc.existLFNsBKRepNo)
            if not res["OK"]:
                gLogger.fatal("Unable to get file metadata", res["Message"])
                return
            if res["Value"]["Failed"]:
                gLogger.error("No metadata found for some files", f"{len(res['Value']['Failed'])} files")
            success = res["Value"]["Successful"]
            filesInvisible = {lfn for lfn, meta in success.items() if meta["VisibilityFlag"][0].upper() == "N"}
            filesVisible = set(success) - filesInvisible
            gLogger.notice(f"{len(filesVisible)} files are visible, {len(filesInvisible)} files are invisible")
            # Try and print the whole as INFO (in case --Verbose was used).
            #   If nothing printed, print a limited number of files as ERROR
            if not gLogger.info(
                "\n".join(
                    f"{lfn} : Visi {success.get(lfn, {}).get('VisibilityFlag', '?')}"
                    for lfn in sorted(cc.existLFNsBKRepNo)
                )
            ):
                if len(cc.existLFNsBKRepNo) > maxFiles:
                    gLogger.notice("First %d files:" % maxFiles)
                gLogger.error(
                    "\n".join(
                        f"{lfn} : Visi {success.get(lfn, {}).get('VisibilityFlag', '?')}"
                        for lfn in sorted(cc.existLFNsBKRepNo)[0:maxFiles]
                    )
                )
            if listAffectedRuns:
                gLogger.notice(f"Affected runs: {','.join(affectedRuns)}")
            gLogger.notice(f"Full list of files:    grep InFCButBKNo {fileName}")
            if fixBK:
                gLogger.notice("Going to fix them, setting the replica flag")
                res = cc.bkClient.addFiles(list(success))
                if res["OK"]:
                    gLogger.notice(f"\tSuccessfully added replica flag to {len(success)} files")
                else:
                    gLogger.error("Failed to set the replica flag", res["Message"])
            elif fixFC:
                gLogger.notice("Going to fix them, by removing from the FC and storage")
                __removeFile(success)
            else:
                gLogger.notice(
                    "Use --FixBK to fix it (set the replica flag) or --FixFC (for removing from FC and storage)"
                )
        else:
            gLogger.notice("====== Completed, no files in the FC with replica = NO in BK ======")
        gLogger.notice("<<<<")

    else:
        gLogger.notice("No files in FC with replica = NO in BK -> OK!")

    if cc.existLFNsNotInBK:
        gLogger.notice(">>>>")

        gLogger.error(f"{len(cc.existLFNsNotInBK)} files are in the FC but are NOT in BK:")
        if fp is None:
            fp = open(fileName, "w")
        fp.write("\nInFCNotInBK ".join([""] + sorted(cc.existLFNsNotInBK)))
        if not gLogger.info("\n".join(sorted(cc.existLFNsNotInBK))):
            if len(cc.existLFNsNotInBK) > maxFiles:
                gLogger.notice("First %d files:" % maxFiles)
            gLogger.error("\n".join(sorted(cc.existLFNsNotInBK[0:maxFiles])))
        gLogger.notice(f"Full list of files:    grep InFCNotInBK {fileName}")
        if fixFC:
            gLogger.notice("Going to fix them, by removing from the FC and storage")
            __removeFile(cc.existLFNsNotInBK)
        else:
            gLogger.notice("Use --FixFC to fix it (remove from FC and storage)")
        gLogger.notice("<<<<")

    else:
        gLogger.notice("No files in FC not in BK -> OK!")
    if fp is not None:
        fp.close()


def doCheckBK2FC(cc, checkAll=False, fixIt=False):
    """Method actually calling for the the check using ConsistencyChecks module
    It prints out results and calls corrective actions if required."""
    cc.checkBK2FC(checkAll)
    maxPrint = 10
    chunkSize = 100
    suffix = ""
    nb = 0
    baseName = "CheckBK2FC" + (f"-{cc.prod}" if cc.prod else "")
    while True:
        fileName = baseName + f"{suffix}.txt"
        if not os.path.exists(fileName):
            break
        nb += 1
        suffix = "-%d" % nb
    fp = None

    if checkAll:
        if cc.existLFNsBKRepNo:
            gLogger.notice(">>>>")

            nFiles = len(cc.existLFNsBKRepNo)
            comment = "%d files are in the FC but have replica = NO in BK" % nFiles
            if nFiles > maxPrint:
                comment += " (first %d LFNs) : \n" % maxPrint
            comment += "\n".join(cc.existLFNsBKRepNo[:maxPrint])
            gLogger.error(comment)
            if fp is None:
                fp = open(fileName, "w")
            fp.write("\nInFCButBKNo ".join([""] + sorted(cc.existLFNsBKRepNo)))
            gLogger.notice("To get the full list:")
            gLogger.notice(f"    grep InFCButBKNo {fileName}")
            if fixIt:
                gLogger.notice("Setting the replica flag...")
                nFiles = 0
                for lfnChunk in breakListIntoChunks(cc.existLFNsBKRepNo, chunkSize):
                    res = cc.bkClient.addFiles(lfnChunk)
                    if not res["OK"]:
                        gLogger.error(f"Something wrong: {res['Message']}")
                    else:
                        nFiles += len(lfnChunk)
                gLogger.notice("Successfully set replica flag to %d files" % nFiles)
            else:
                gLogger.notice("Use option --FixIt to fix it (set the replica flag)")
            gLogger.notice("<<<<")

        else:
            gLogger.notice("No LFNs exist in the FC but have replicaFlag = No in the BK -> OK!")

    if cc.absentLFNsBKRepYes:
        gLogger.notice(">>>>")

        nFiles = len(cc.absentLFNsBKRepYes)
        comment = "%d files have replicaFlag = Yes but are not in FC:" % nFiles
        if nFiles > maxPrint:
            comment += " (first %d LFNs) : \n" % maxPrint
        comment += "\n".join(cc.absentLFNsBKRepYes[:maxPrint])
        gLogger.error(comment)
        if fp is None:
            fp = open(fileName, "w")
        fp.write("\nInBKButNotInFC ".join([""] + sorted(cc.absentLFNsBKRepYes)))
        gLogger.notice("To get the full list:")
        gLogger.notice(f"    grep InBKButNotInFC {fileName}")
        if fixIt:
            gLogger.notice("Removing the replica flag...")
            nFiles = 0
            for lfnChunk in breakListIntoChunks(cc.absentLFNsBKRepYes, chunkSize):
                res = cc.bkClient.removeFiles(lfnChunk)
                if not res["OK"]:
                    gLogger.error("Something wrong:", res["Message"])
                else:
                    nFiles += len(lfnChunk)
            gLogger.notice("Successfully removed replica flag to %d files" % nFiles)
        else:
            gLogger.notice("Use option --FixIt to fix it (remove the replica flag)")
        gLogger.notice("<<<<")

    else:
        gLogger.notice("No LFNs have replicaFlag = Yes but are not in the FC -> OK!")
    if fp:
        fp.close()


def doCheckSE(cc, seList, fixIt=False):
    cc.checkSE(seList)

    if cc.absentLFNsInFC:
        gLogger.notice(">>>>")
        gLogger.notice(f"{len(cc.absentLFNsInFC)} files are not in the FC")
        if fixIt:
            __removeFile(cc.absentLFNsInFC)
        else:
            gLogger.notice("Use --FixIt to fix it (remove from catalogs")
        gLogger.notice("<<<<")

    if cc.existLFNsNoSE:
        gLogger.notice("<<<<")
        gLogger.notice(f"{len(cc.existLFNsNoSE)} files are not present at {', '.join(sorted(seList))}")
        gLogger.notice("\n".join(sorted(cc.existLFNsNoSE)))
    else:
        gLogger.notice(f"No LFNs missing at {', '.join(sorted(seList))}")



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Returns descendants for a (list of) LFN(s)"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "All", "Do not restrict to descendants with replicas")
    Script.registerSwitch("", "Full", "Get full metadata information on descendants")
    level = 1
    Script.registerSwitch("", "Depth=", "Number of processing levels (default: %d)" % level)
    Script.registerSwitch("", "Production=", "Restrict to descendants in a given production (at any depth)")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... [LFN|File] [Depth]",
                "Arguments:",
                "  LFN:      Logical File Name",
                "  File:     Name of the file with a list of LFNs",
                "  Depth:    Number of levels to search (default: %d)" % level,
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeFileDescendants

    executeFileDescendants(dmScript, level)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbScript is very similar to DIRAC Script module, but consider LHCb
environment."""
import os

from DIRAC import gLogger
from DIRAC.Core.Utilities import DErrno
from DIRAC.Workflow.Modules.Script import Script


class LHCbScript(Script):
    """A simple extension to the DIRAC script module."""

    def __init__(self):
        """c'tor."""
        self.log = gLogger.getSubLogger("LHCbScript")
        super().__init__(self.log)

        self.systemConfig = "ANY"
        self.environment = {}

    def _resolveInputVariables(self):
        """By convention the workflow parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.systemConfig = self.step_commons.get("SystemConfig", self.systemConfig)

    def _executeCommand(self):
        """Executes the self.command (uses systemCall) with binary tag (CMTCONFIG)
        requested (if not 'ANY')"""

        if self.systemConfig != "ANY":
            self.environment = os.environ
            self.environment["CMTCONFIG"] = self.systemConfig

        super()._executeCommand()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""unit tests for Configuration Helpers."""
from unittest import mock
import os
import pytest

from DIRAC import gLogger

from DIRAC.Core.Security.Locations import getCAsLocation
import LHCbDIRAC.ConfigurationSystem.Client.Helpers.Resources as moduleTested

gLogger.setLevel("DEBUG")


@pytest.mark.parametrize(
    "binaryTagsSet, expected",
    [
        (set(), None),
        ({frozenset({"x86_64-slc6-gcc48-opt"})}, "x86_64-slc6"),
        ({frozenset({"x86_64-slc6-gcc62-opt"})}, "x86_64_v2-slc6"),
        ({frozenset({"x86_64+avx2-slc6-gcc62-opt"})}, "x86_64_v3-slc6"),
        ({frozenset({"x86_64+avx2+fma-centos7-gcc7-opt"})}, "x86_64_v3-centos7"),
        ({frozenset({"x86_64-slc6-gcc48-opt"}), frozenset({"x86_64-slc6-gcc62-opt"})}, "x86_64_v2-slc6"),
        ({frozenset({"x86_64-centos7-gcc49-opt"}), frozenset({"x86_64-slc6-gcc62-opt"})}, "x86_64_v2-any"),
        (
            {
                frozenset({"x86_64-centos7-gcc49-opt"}),
                frozenset({"x86_64-slc6-gcc62-opt"}),
                frozenset({"x86_64+avx2-centos7-gcc62-opt"}),
            },
            "x86_64_v3-any",
        ),
        (
            {
                frozenset({"x86_64-centos7-gcc49-opt"}),
                frozenset({"x86_64+avx-slc6-gcc62-opt"}),
                frozenset({"x86_64+avx2-centos7-gcc62-opt"}),
            },
            "x86_64_v3-any",
        ),
        (
            {
                frozenset({"x86_64+fma-centos7-gcc49-opt"}),
                frozenset({"x86_64+avx-slc6-gcc62-opt"}),
                frozenset({"x86_64+avx2-centos7-gcc62-opt"}),
            },
            "x86_64_v3-any",
        ),
        (
            {
                frozenset({"x86_64+avx2+fma-centos7-gcc49-opt"}),
                frozenset({"x86_64+avx-slc6-gcc62-opt"}),
                frozenset({"x86_64+avx2+fma-centos7-gcc62-opt"}),
            },
            "x86_64_v3-any",
        ),
        (
            {
                frozenset({"x86_64+avx2+fma-centos7-gcc49-opt"}),
                frozenset({"x86_64+avx-slc6-gcc7-opt"}),
                frozenset({"x86_64+avx2+fma-centos7-gcc62-opt"}),
            },
            "x86_64_v3-any",
        ),
        ({frozenset({"x86_64+fma-centos7-gcc7-opt"}), frozenset({"x86_64+avx-slc6-gcc62-opt"})}, "x86_64_v3-any"),
        ({frozenset({"x86_64-slc5-gcc49-opt"}), frozenset({"x86_64+fma-centos7-gcc7-opt"})}, "x86_64_v3-any"),
        (
            {
                frozenset({"x86_64-slc5-gcc49-opt"}),
                frozenset({"x86_64+fma-centos7-gcc7-opt", "x86_64+fma-slc6-gcc7-opt"}),
            },
            "x86_64_v3-slc6",
        ),
        (
            {
                frozenset({"x86_64-slc5-gcc49-opt", "x86_64-slc6-gcc49-opt"}),
                frozenset({"x86_64+fma-centos7-gcc7-opt", "x86_64-slc6-gcc7-opt"}),
            },
            "x86_64_v2-slc6",
        ),
    ],
)
def test_getPlatformForJob(binaryTagsSet, expected):
    fbtMock = mock.MagicMock()
    fbtMock.return_value = binaryTagsSet
    moduleTested._findBinaryTags = fbtMock

    res = moduleTested.getPlatformForJob(mock.MagicMock())
    assert res == expected


@pytest.mark.parametrize(
    "platform, expectedRes, expectedValue",
    [
        ("", True, []),
        ("ANY", True, []),
        (["ANY"], True, []),
        ([None], True, []),
        (["bih", "boh"], False, []),
        (
            "x86_64-slc6",
            True,
            [
                "cannonlake-slc6",
                "skylake_avx512-slc6",
                "skylake-slc6",
                "broadwell-slc6",
                "haswell-slc6",
                "ivybridge-slc6",
                "sandybridge-slc6",
                "westmere-slc6",
                "nehalem-slc6",
                "core2-slc6",
                "x86_64-slc6",
            ],
        ),
        (
            "x86_64-centos7",
            True,
            [
                "cannonlake-centos7",
                "skylake_avx512-centos7",
                "skylake-centos7",
                "broadwell-centos7",
                "haswell-centos7",
                "ivybridge-centos7",
                "sandybridge-centos7",
                "westmere-centos7",
                "nehalem-centos7",
                "core2-centos7",
                "x86_64-centos7",
            ],
        ),
        (
            "x86_64-slc5",
            True,
            [
                "cannonlake-slc6",
                "skylake_avx512-slc6",
                "skylake-slc6",
                "broadwell-slc6",
                "haswell-slc6",
                "ivybridge-slc6",
                "sandybridge-slc6",
                "westmere-slc6",
                "nehalem-slc6",
                "core2-slc6",
                "x86_64-slc6",
                "cannonlake-slc5",
                "skylake_avx512-slc5",
                "skylake-slc5",
                "broadwell-slc5",
                "haswell-slc5",
                "ivybridge-slc5",
                "sandybridge-slc5",
                "westmere-slc5",
                "nehalem-slc5",
                "core2-slc5",
                "x86_64-slc5",
            ],
        ),
        (
            "x86_64-slc5.avx2",
            True,
            [
                "cannonlake-slc6",
                "skylake_avx512-slc6",
                "skylake-slc6",
                "broadwell-slc6",
                "haswell-slc6",
                "cannonlake-slc5",
                "skylake_avx512-slc5",
                "skylake-slc5",
                "broadwell-slc5",
                "haswell-slc5",
            ],
        ),
        (
            "x86_64-slc6.avx2+fma",
            True,
            [
                "cannonlake-slc6",
                "skylake_avx512-slc6",
                "skylake-slc6",
                "broadwell-slc6",
                "haswell-slc6",
            ],
        ),
    ],
)
def test_getDIRACPlatform(platform, expectedRes, expectedValue):
    res = moduleTested.getDIRACPlatform(platform)
    assert res["OK"] is expectedRes
    if res["OK"]:
        assert set(expectedValue).issubset(set(res["Value"])) is True


@pytest.mark.parametrize(
    "applicationName, applicationVersion, expected",
    [
        (
            "DaVinci",
            "v44r4",
            [
                "x86_64+avx2+fma-centos7-gcc62-opt",
                "x86_64-centos7-gcc62-dbg",
                "x86_64-centos7-gcc62-do0",
                "x86_64-centos7-gcc62-opt",
                "x86_64-centos7-gcc7-dbg",
                "x86_64-centos7-gcc7-opt",
                "x86_64-slc6-gcc62-dbg",
                "x86_64-slc6-gcc62-opt",
            ],
        ),
        ("DaVinci", "v0r0", None),
        ("InvalidProject", "v1r0", None),
        (
            "CASTELAO",
            "v1r5p1",
            [
                "x86_64-centos7-gcc62-do0",
                "x86_64-centos7-gcc62-opt",
                "x86_64-centos7-gcc62-dbg",
                "x86_64-slc6-gcc62-opt",
                "x86_64-centos7-gcc7-dbg",
                "x86_64+avx2+fma-centos7-gcc62-opt",
                "x86_64-slc6-gcc62-dbg",
                "x86_64-centos7-gcc7-opt",
            ],
        ),
    ],
)
def test_listPlatforms(applicationName, applicationVersion, expected):
    if not os.path.isdir("/cvmfs/lhcb.cern.ch"):
        pytest.skip("CVMFS is required")

    # Good cache path
    result = moduleTested._listPlatforms(applicationName, applicationVersion, moduleTested.DEFAULT_CACHEPATH)
    assert result is None is expected or set(result) == set(expected)

    # Invalid cache path
    result = moduleTested._listPlatforms(
        applicationName, applicationVersion, "/cvmfs/lhcb.cern.invalid/lib/var/lib/softmetadata/project-platforms.json"
    )
    assert result is None



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to upload specified job output files according to the parameters
defined in the production workflow."""
import os
import random
import glob
from operator import itemgetter

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import DEncode
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from LHCbDIRAC.Core.Utilities.ResolveSE import getDestinationSEList
from LHCbDIRAC.Core.Utilities.ProductionData import constructProductionLFNs
from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import getFileDescendents

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class UploadOutputData(ModuleBase):
    """Module to upload specified job output files according to the parameters
    defined in the production workflow."""

    #############################################################################
    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("UploadOutputData")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.commandTimeOut = 10 * 60
        self.jobID = ""
        self.existingCatalogs = []
        result = gConfig.getSections("/Resources/FileCatalogs")
        if result["OK"]:
            self.existingCatalogs = result["Value"]

        # List all parameters here
        self.outputDataFileMask = ""
        self.outputMode = "Any"  # or 'Run', for Reco/Stripping case, or 'Local', e.g. for MCMerge
        self.outputList = []
        self.outputDataStep = ""
        self.request = None
        self.failoverTransfer = None
        self.prodOutputLFNs = []
        self.failoverSEs = None

    #############################################################################
    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""

        super()._resolveInputVariables()

        if "outputDataStep" in self.workflow_commons:
            self.outputDataStep = [str(ds) for ds in self.workflow_commons["outputDataStep"].split(";")]

        if "outputList" in self.workflow_commons:
            self.outputList = self.workflow_commons["outputList"]

        if "outputMode" in self.workflow_commons:
            self.outputMode = self.workflow_commons["outputMode"]

        # Use LHCb utility for local running via jobexec
        if "ProductionOutputData" in self.workflow_commons:
            self.prodOutputLFNs = self.workflow_commons["ProductionOutputData"]
            if isinstance(self.prodOutputLFNs, str):
                self.prodOutputLFNs = [i.strip() for i in self.prodOutputLFNs.split(";")]  # pylint: disable=no-member
        else:
            self.log.info("ProductionOutputData parameter not found, creating on the fly")
            result = constructProductionLFNs(self.workflow_commons, self.bkClient)
            if not result["OK"]:
                self.log.error("Could not create production LFNs", result["Message"])
                return result
            self.prodOutputLFNs = result["Value"]["ProductionOutputData"]

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
        SEs=None,
        fileDescendants=None,
    ):
        """Main execution function.

        1. Determine the final list of possible output files for the workflow
           and all the parameters needed to upload them.
        2. Verifying that the input files have no descendants (and exiting with error, otherwise)
        3. Sending the BK records for the steps of the job
        4. Transfer output files in their destination, register in the FC (with failover)
        5. Registering the output files in the Bookkeeping
        """

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            # This returns all Tier1-Failover unless a specific one is defined for the site
            self.failoverSEs = getDestinationSEList("Tier1-Failover", self.siteName, outputmode="Any")
            random.shuffle(self.failoverSEs)

            self._resolveInputVariables()

            if not self._checkWFAndStepStatus():
                return S_OK("Failures detected in previous steps: no output data upload attempted")

            # ## 1. Determine the final list of possible output files
            # ##    for the workflow and all the parameters needed to upload them.
            # ##

            self.log.verbose("Getting the list of candidate files")
            fileDict = self.getCandidateFiles(
                self.outputList, self.prodOutputLFNs, self.outputDataFileMask, self.outputDataStep
            )

            fileMetadata = self.getFileMetadata(fileDict)

            if not fileMetadata:
                self.log.info("No output data files were determined to be uploaded for this workflow")
                return S_OK()

            # Get final, resolved SE list for files
            final = {}

            for fileName, metadata in fileMetadata.items():
                if not SEs:
                    resolvedSE = getDestinationSEList(
                        metadata["workflowSE"], self.siteName, self.outputMode, self.workflow_commons.get("runNumber")
                    )
                else:
                    resolvedSE = SEs
                final[fileName] = metadata
                final[fileName]["resolvedSE"] = resolvedSE

            self.log.info("The following files will be uploaded", f": {', '.join(final)}")
            for fileName, metadata in final.items():
                self.log.info(f"--------{fileName}--------")
                for name, val in metadata.items():
                    self.log.info(f"{name} = {val}")

            if not self._enableModule():
                # At this point can exit and see exactly what the module would have uploaded
                self.log.info("Module disabled", f"would have attempted to upload the files {', '.join(final)}")

            # ## 2. Prior to uploading any files must check (for productions with input data) that no descendant files
            # ##    already exist with replica flag in the BK.
            # ##

            if self.inputDataList:
                if fileDescendants is not None:
                    lfnsWithDescendants = fileDescendants
                else:
                    if not self._enableModule():
                        self.log.info(
                            "Module disabled",
                            f"would have attempted to check the files {', '.join(self.inputDataList)}",
                        )
                        lfnsWithDescendants = []
                    else:
                        lfnsWithDescendants = getFileDescendents(
                            self.production_id, self.inputDataList, dm=self.dataManager, bkClient=self.bkClient
                        )
                if not lfnsWithDescendants:
                    self.log.info("No descendants found, outputs can be uploaded")
                else:
                    self.log.error("Found descendants!!! Outputs won't be uploaded")
                    self.log.info("Files with descendants", ": %s" " % ".join(lfnsWithDescendants))
                    self.log.info(
                        "The files above will be set as 'Processed', other lfns in input will be later reset as Unused"
                    )
                    self.fileReport.setFileStatus(int(self.production_id), lfnsWithDescendants, "Processed")
                    return S_ERROR("Input Data Already Processed")

            # ## 3. Sending the BK records for the steps of the job
            # ##

            bkFileExtensions = ["bookkeeping*.xml"]
            bkFiles = []
            for ext in bkFileExtensions:
                self.log.debug(f"Looking at BK record wildcard: {ext}")
                globList = glob.glob(ext)
                for check in globList:
                    if os.path.isfile(check):
                        self.log.verbose("Found locally existing BK file record", f": {check}")
                        bkFiles.append(check)

            # Unfortunately we depend on the file names to order the BK records
            bkFilesListTuples = []
            for bk in bkFiles:
                bkFilesListTuples.append((bk, int(bk.split("_")[-1].split(".")[0])))
            bkFiles = [bk[0] for bk in sorted(bkFilesListTuples, key=itemgetter(1))]

            self.log.info("The following BK records will be sent", f": {', '.join(bkFiles)}")
            for bkFile in bkFiles:
                with open(bkFile) as fd:
                    bkXML = fd.read()
                self.log.info("Sending BK record", f":\n{bkXML}")
                if self._enableModule():
                    result = self.bkClient.sendXMLBookkeepingReport(bkXML)
                    self.log.verbose(result)
                    if result["OK"]:
                        self.log.info("Bookkeeping report sent", f"for {bkFile}")
                    else:
                        self.log.error("Could not send Bookkeeping XML file to server", f": {result['Message']}")
                        self.log.info("Preparing DISET request", f"for {bkFile}")
                        bkDISETReq = Operation()
                        bkDISETReq.Type = "ForwardDISET"
                        bkDISETReq.Arguments = DEncode.encode(result["rpcStub"])
                        self.request.addOperation(bkDISETReq)
                        self.workflow_commons["Request"] = self.request  # update each time, just in case
                else:
                    self.log.info("Would have attempted to send a bk record, but the module is disabled")

            # ## 4. Transfer output files in their destination, register in the FC (with failover)
            # ##

            # Disable the watchdog check in case the file uploading takes a long time
            self._disableWatchdogCPUCheck()

            # Instantiate the failover transfer client with the global request object
            if not self.failoverTransfer:
                self.failoverTransfer = FailoverTransfer(self.request)

            # Track which files are successfully uploaded (not to failover) via
            performBKRegistration = []
            # Failover replicas are always added to the BK when they become available (actually, added to all the catalogs)

            failover = {}
            for fileName, metadata in final.items():
                targetSE = metadata["resolvedSE"]
                self.log.info(
                    "Attempting to store file to SE", f"{fileName} to the following SE(s):\n{', '.join(targetSE)}"
                )
                fileMetaDict = {
                    "Size": metadata["filedict"]["Size"],
                    "LFN": metadata["filedict"]["LFN"],
                    "GUID": metadata["filedict"]["GUID"],
                    "Checksum": metadata["filedict"]["Checksum"],
                    "ChecksumType": metadata["filedict"]["ChecksumType"],
                }

                if not self._enableModule():
                    # At this point can exit and see exactly what the module would have uploaded
                    self.log.info("Module disabled", f"would have attempted to upload file {fileName}")
                    continue

                result = self.failoverTransfer.transferAndRegisterFile(
                    fileName=fileName,
                    localPath=metadata["localpath"],
                    lfn=metadata["filedict"]["LFN"],
                    destinationSEList=targetSE,
                    fileMetaDict=fileMetaDict,
                    masterCatalogOnly=True,
                )
                if not result["OK"]:
                    self.log.error("Could not transfer and register", f" {fileName} with metadata:\n {metadata}")
                    failover[fileName] = metadata
                else:
                    self.log.info(
                        "File uploaded, will be registered in BK if all files uploaded for job", f"({fileName})"
                    )

                    # if the files are uploaded in the SE, independently if the registration in the FC is done,
                    # then we have to register all of them in the BKK
                    performBKRegistration.append(metadata)

            cleanUp = False
            for fileName, metadata in failover.items():
                self.log.info("Setting default catalog for failover transfer registration to master catalog")
                random.shuffle(self.failoverSEs)
                targetSE = metadata["resolvedSE"][0]
                metadata["resolvedSE"] = self.failoverSEs

                fileMetaDict = {
                    "Size": metadata["filedict"]["Size"],
                    "LFN": metadata["filedict"]["LFN"],
                    "GUID": metadata["filedict"]["GUID"],
                    "Checksum": metadata["filedict"]["Checksum"],
                    "ChecksumType": metadata["filedict"]["ChecksumType"],
                }

                if not self._enableModule():
                    # At this point can exit and see exactly what the module would have uploaded
                    self.log.info("Module disabled", f"would have attempted to upload with failover file {fileName}")
                    continue

                result = self.failoverTransfer.transferAndRegisterFileFailover(
                    fileName=fileName,
                    localPath=metadata["localpath"],
                    lfn=metadata["filedict"]["LFN"],
                    targetSE=targetSE,
                    failoverSEList=metadata["resolvedSE"],
                    fileMetaDict=fileMetaDict,
                    masterCatalogOnly=True,
                )
                if not result["OK"]:
                    self.log.error(
                        "Could not transfer and register", f"{fileName} in failover with metadata:\n {metadata}"
                    )
                    cleanUp = True
                    break  # no point continuing if one completely fails

            # Now after all operations, retrieve potentially modified request object
            self.request = self.failoverTransfer.request

            # If some or all of the files failed to be saved even to failover
            if cleanUp and self._enableModule():
                self._cleanUp(final)
                self.workflow_commons["Request"] = self.request
                return S_ERROR("Failed to upload output data")

            # For files correctly uploaded must report LFNs to job parameters
            if final and self._enableModule():
                report = ", ".join(final)
                self.setJobParameter("UploadedOutputData", report)

            # ## 5. Can now register the successfully uploaded files in the BK i.e. set the BK replica flags
            # ##

            if not performBKRegistration:
                self.log.info("There are no files to perform the BK registration for, all are in failover")
            elif self._enableModule():
                # performing BK registration

                # Getting what should be registered immediately, and what later
                lfnsToRegisterInBK = {metadata["filedict"]["LFN"] for metadata in performBKRegistration}
                lfnsToRegisterInBKNow = self._getLFNsForBKRegistration(lfnsToRegisterInBK)
                lfnsToRegisterInBKLater = list(lfnsToRegisterInBK - set(lfnsToRegisterInBKNow))

                # Registering what should be registering immediately, and handling failures
                if lfnsToRegisterInBKNow:
                    result = FileCatalog(catalogs=["BookkeepingDB"]).addFile(lfnsToRegisterInBKNow)
                    self.log.verbose(f"BookkeepingDB.addFile: {result}")
                    if not result["OK"]:
                        self.log.error(result)
                        return S_ERROR("Could Not Perform BK Registration")
                    if "Failed" in result["Value"] and result["Value"]["Failed"]:
                        for lfn, error in result["Value"]["Failed"].items():
                            lfnMetadata = {}
                            for lfnMD in performBKRegistration:
                                if (
                                    lfnMD["lfn"] == lfn
                                ):  # the lfn is indeed both at lfnMD['lfn'] and at lfnMD['filedict']['LFN']
                                    lfnMetadata = lfnMD["filedict"]
                                    break
                            self.setBKRegistrationRequest(lfn, error=error, metaData=lfnMetadata)

                # Adding a registration request for what whould be registered later
                if lfnsToRegisterInBKLater:
                    for lfnMD in performBKRegistration:
                        if lfnMD["lfn"] in lfnsToRegisterInBKLater:
                            lfnMetadata = lfnMD["filedict"]
                            self.setBKRegistrationRequest(lfnMD["lfn"], metaData=lfnMetadata)

            self.workflow_commons["Request"] = self.request

            return S_OK("Output data uploaded")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Exception in UploadOutputData", lException=e)
            self.setApplicationStatus(repr(e))
            return S_ERROR(str(e))

        finally:
            super().finalize()

    #############################################################################

    def _getLFNsForBKRegistration(self, lfns):
        """Check what should be registered immediately in the BK, and what later.
        If there's a request in self.request for registering the file in the FC,
        don't perform the registration in the BK immediately: in this case the file
        should be registered with an operation.

        :param list lfnsList: an iterable of LFNs
        :return: list of LFNs to be registered immediately
        """

        postPonePerformBKRegistration = []
        for op in self.request:
            if op.Type == "RegisterFile":  # We assume that this is a registerFile operation for the DFC... what else?
                for fileInOp in op:
                    if fileInOp.LFN in lfns:
                        postPonePerformBKRegistration.append(fileInOp.LFN)

        return list(set(lfns) - set(postPonePerformBKRegistration))

    def _cleanUp(self, final):
        """Clean up uploaded data for the LFNs in the list."""
        lfnList = []
        for _fileName, metadata in final.items():
            lfnList.append(metadata["lfn"])

        self.log.verbose(f"Cleaning up the request, for LFNs: {', '.join(lfnList)}")

        newRequest = Request()

        for op in self.request:
            add = True
            if op.Type in [
                "PutAndRegister",
                "ReplicateAndRegister",
                "RegisterFile",
                "RegisterReplica",
                "RemoveReplica",
            ]:
                for files in op:
                    if files.LFN in lfnList:
                        add = False
            if add:
                newRequest.addOperation(op)

        self.request = newRequest

        self.log.verbose(f"And adding RemoveFile operation for LFNs: {', '.join(lfnList)}, just in case")

        removeFiles = Operation()
        removeFiles.Type = "RemoveFile"
        for lfn in lfnList:
            removedFile = File()
            removedFile.LFN = lfn
            removeFiles.addFile(removedFile)
        self.request.addOperation(removeFiles)


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Run a local test job for based on a production request YAML specification"""
from __future__ import annotations

import json
import os
import random
import shlex
import subprocess
import sys
import tempfile
from collections import defaultdict
from pathlib import Path
from textwrap import dedent

import yaml

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue


from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import (
    parse_obj,
    ProductionBase,
    SimulationProduction,
    DataProduction,
    ProductionStep,
)


def preprocess_spec(pr: ProductionBase):
    """production-request-run-local does not support multiple input types so easily, so we need to split them"""
    new_steps = []

    for step in pr.steps:
        if len(step.input) > 1:
            print(f"Splitting step {step.name}")
            # we need to split this step
            for i, input_filetype in enumerate(step.input):
                split_step = step.model_copy(
                    deep=True,
                    update={
                        "name": f"{step.name}ft{i}",
                        "processing_pass": f"{step.processing_pass}{i}",
                        "input": [input_filetype],
                        "output": [input_filetype.model_copy(deep=True)],
                    },
                )
                new_steps.append(split_step)
                print(split_step)
        else:
            new_steps.append(step)
    return pr.model_copy(
        deep=True,
        update={
            "steps": new_steps,
        },
    )


def parseArgs():
    useCfgOverride = True
    inputFiles = None
    inputFileType = None
    ancestorDepth = None
    numTestLFNs = 1
    testRunNumbers = None
    startRun = None
    endRun = None
    exportTestLFNs = None

    @convertToReturnValue
    def disableCfgOverride(_):
        nonlocal useCfgOverride
        useCfgOverride = False

    @convertToReturnValue
    def setInputFiles(s: str):
        nonlocal inputFiles
        inputFiles = s.split(",")

    @convertToReturnValue
    def setOutputFileType(s: str):
        nonlocal inputFileType
        inputFileType = s

    @convertToReturnValue
    def setAncestorDepth(s: str):
        nonlocal ancestorDepth
        ancestorDepth = int(s)

    @convertToReturnValue
    def setNumTestLFNs(s: str):
        nonlocal numTestLFNs
        numTestLFNs = int(s)

    @convertToReturnValue
    def setTestRunNumbers(s: str):
        nonlocal testRunNumbers
        testRunNumbers = s.split(",")

    @convertToReturnValue
    def setStartRun(s: str):
        nonlocal startRun
        startRun = int(s)

    @convertToReturnValue
    def setEndRun(s: str):
        nonlocal endRun
        endRun = int(s)

    @convertToReturnValue
    def setExportTestLFNsPath(s: str):
        nonlocal exportTestLFNs
        exportTestLFNs = s

    switches = [
        ("", "input-files=", "Comma separated list of input files (Data only)", setInputFiles),
        ("", "no-cfg-override", "Internal implementation detail", disableCfgOverride),
        ("", "input-file-type=", "Limit the file type for generic merge steps", setOutputFileType),
        ("", "ancestor-depth=", "Set the ancestor depth that should be included in the pool catalog", setAncestorDepth),
        ("", "num-test-lfns=", "Number of LFNs to test with", setNumTestLFNs),
        ("", "test-runs=", "Comma separated list of test runs to use", setTestRunNumbers),
        ("", "start-run=", "Start run number for data production", setStartRun),
        ("", "end-run=", "End run number for data production", setEndRun),
        (
            "",
            "export-test-lfns=",
            "Export the LFNs that would have been used for testing to a file",
            setExportTestLFNsPath,
        ),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("yaml_path: Path to the YAML file containing productions to submit")
    Script.registerArgument("name: Name of the production to submit", mandatory=False)
    Script.registerArgument("event_type: The event type to generate (Simulation only)", mandatory=False)
    Script.parseCommandLine(ignoreErrors=False)
    yaml_path, name, eventType = Script.getPositionalArgs(group=True)

    from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient

    if not ConfigurationClient().ping()["OK"]:
        gLogger.fatal("Failed to contact CS, do you have a valid proxy?")
        sys.exit(1)

    if (startRun or endRun) and testRunNumbers:
        gLogger.fatal("Cannot specify both --test-runs and --start-run/--end-run")
        sys.exit(1)

    return (
        Path(yaml_path),
        name,
        eventType,
        inputFiles,
        useCfgOverride,
        inputFileType,
        ancestorDepth,
        numTestLFNs,
        testRunNumbers,
        startRun,
        endRun,
        exportTestLFNs,
    )


def _runWithConfigOverride(argv):
    """Relaunch the process with DIRACSYSCONFIG overridden for local tests"""
    cfg_content = f"""
    DIRAC
    {{
        Setup={CSGlobals.getSetup()}
    }}
    LocalSite
    {{
        Site = DIRAC.LocalProdTest.local
        GridCE = jenkins.cern.ch
        CEQueue = jenkins-queue_not_important
        LocalSE = CERN-DST-EOS
        LocalSE += CERN-HIST-EOS
        LocalSE += CERN-RAW
        LocalSE += CERN-FREEZER-EOS
        LocalSE += CERN-SWTEST
        Architecture = x86_64-centos7
        SharedArea = /cvmfs/lhcb.cern.ch/lib
        CPUTimeLeft = 123456
    }}
    """
    with tempfile.NamedTemporaryFile(mode="wt") as tmp:
        tmp.write(dedent(cfg_content))
        tmp.flush()

        env = dict(os.environ)
        env["DIRACSYSCONFIG"] = ",".join([tmp.name] + env.get("DIRACSYSCONFIG", "").split(","))

        gLogger.always("Overriding DIRACSYSCONFIG to", env["DIRACSYSCONFIG"])
        gLogger.always("Restarting process with", argv)
        proc = subprocess.run(argv, env=env, check=False)
    sys.exit(proc.returncode)


@Script()
def main():
    (
        yamlPath,
        name,
        eventType,
        inputFiles,
        useCfgOverride,
        inputFileType,
        ancestorDepth,
        numTestLFNs,
        testRunNumbers,
        startRun,
        endRun,
        exportTestLFNs,
    ) = parseArgs()

    if useCfgOverride:
        return _runWithConfigOverride(sys.argv + ["--no-cfg-override"])

    productionRequests = defaultdict(list)
    for spec in yaml.safe_load(yamlPath.read_text()):
        productionRequest = parse_obj(spec)
        productionRequest = preprocess_spec(productionRequest)
        productionRequests[productionRequest.name] += [productionRequest]

    if name is None:
        if len(productionRequests) == 1:
            name = list(productionRequests)[0]
        else:
            gLogger.fatal(
                "Multiple production requests available, please specify a name. Available options are:\n",
                "   * " + "\n    * ".join(map(shlex.quote, productionRequests)),
            )
            sys.exit(1)
    if name not in productionRequests:
        gLogger.fatal(
            "Unrecognised production request name. Available options are:\n",
            "   * " + "\n    * ".join(map(shlex.quote, productionRequests)),
        )
        sys.exit(1)
    if len(productionRequests[name]) > 1:
        gLogger.fatal("Ambiguous production requests found with identical names", shlex.quote(name))
        sys.exit(1)
    productionRequest = productionRequests[name][0]

    numTestEvents = None
    if isinstance(productionRequest, SimulationProduction):
        availableEventTypes = {e.id: e.num_test_events for e in productionRequest.event_types}
        if eventType is None and isinstance(productionRequest, SimulationProduction):
            if len(productionRequest.event_types) == 1:
                eventType = productionRequest.event_types[0].id
            else:
                gLogger.fatal(
                    "Multiple event types available, please specify a one.\nAvailable options are:\n",
                    "   * " + "\n    * ".join(availableEventTypes),
                )
                sys.exit(1)
        if eventType not in availableEventTypes:
            gLogger.fatal(f"Invalid event type passed ({eventType}), available options are: {availableEventTypes!r}")
            sys.exit(1)
        numTestEvents = availableEventTypes[eventType]
    elif eventType is not None:
        gLogger.fatal(f"{eventType!r} but this is not a simulation production!")
        sys.exit(1)

    pr, kwargs = prepareProductionRequest(
        productionRequest,
        eventType=eventType,
        numTestEvents=numTestEvents,
        inputFiles=inputFiles,
        inputFileType=inputFileType,
        ancestorDepth=ancestorDepth,
        numTestLFNs=numTestLFNs,
        testRunNumbers=testRunNumbers,
        startRun=startRun,
        endRun=endRun,
    )

    if exportTestLFNs:
        Path(exportTestLFNs).write_text(json.dumps(kwargs["inputDataList"]))
        gLogger.notice(f"Exported test LFNs to {exportTestLFNs}")
        sys.exit(0)

    # TODO: pr._buildProduction eats stepsInProd
    prod = pr._buildProduction(**kwargs)
    returnValueOrRaise(prod.runLocal())


def prepareProductionRequest(
    productionRequest: ProductionBase,
    *,
    eventType: str | None = None,
    numTestEvents: int = 10,
    inputFiles: list[str] | None = None,
    inputFileType: str | None = None,
    ancestorDepth=None,
    numTestLFNs: int = 1,
    testRunNumbers: list[str] | None = None,
    startRun: int | None = None,
    endRun: int | None = None,
):
    """Prepare a ProductionRequest for running locally and return the kwargs to pass to ProductionRequest._buildProduction"""
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import (
        production_to_legacy_dict,
        configure_input,
    )
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

    pr = ProductionRequest()
    kwargs = {}
    legacy_dict, _ = production_to_legacy_dict(productionRequest)
    pr.prodGroup = json.loads(legacy_dict["ProDetail"])["pDsc"]
    if isinstance(productionRequest, SimulationProduction):
        pr.configName = "MC"
        pr.configVersion = productionRequest.mc_config_version
        pr.dataTakingConditions = productionRequest.sim_condition
        pr.eventType = eventType

        kwargs |= dict(
            events=numTestEvents,
            multicore=False,
            prodType="MCSimulation" if productionRequest.fast_simulation_type == "None" else "MCFastSimulation",
        )
    elif isinstance(productionRequest, DataProduction):
        if not inputFiles:
            configure_input(pr, legacy_dict, runs=testRunNumbers, startRun=startRun, endRun=endRun)
            bkQueryDict = pr._getBKKQuery().copy()
            if "RunNumbers" in bkQueryDict:
                bkQueryDict["RunNumbers"] = bkQueryDict["RunNumbers"].split(";;;")
            if "DataQualityFlag" in bkQueryDict:
                bkQueryDict["DataQualityFlag"] = bkQueryDict["DataQualityFlag"].split(";;;")
            if "ExtendedDQOK" in bkQueryDict:
                bkQueryDict["ExtendedDQOK"] = bkQueryDict["ExtendedDQOK"].split(";;;")
            result = returnValueOrRaise(BookkeepingClient().getFilesWithMetadata(bkQueryDict))
            if result["TotalRecords"] == 0:
                raise ValueError("No input files found in the bookkeeping")

            # Remove the smallest 50% of files to avoid unusually small files
            sizeIndex = result["ParameterNames"].index("FileSize")
            records = sorted(result["Records"], key=lambda x: x[sizeIndex])
            if len(records) // 2 >= numTestLFNs:
                records = records[len(records) // 2 :]

            # Shuffle the LFNs so we pick a random one
            random.shuffle(records)

            # Only run tests with files which have available replicas
            filenameIndex = result["ParameterNames"].index("FileName")
            inputFiles = []
            for record in records:
                lfn = record[filenameIndex]
                result = returnValueOrRaise(DataManager().getReplicasForJobs([lfn], diskOnly=True))
                inputFiles.extend(result["Successful"])
                if len(inputFiles) == numTestLFNs:
                    break
                if result["Failed"]:
                    gLogger.warn("Skipping LFN as it has no replicas for jobs", result["Failed"])
            else:
                raise ValueError(f"Only {len(inputFiles)} input files found, requested {numTestLFNs}")

        if len(inputFiles) < numTestLFNs:
            raise ValueError(f"Only {len(inputFiles)} input files found, requested {numTestLFNs}")

        kwargs |= dict(
            inputDataList=inputFiles,
            prodType=productionRequest.type,
            inputDataPolicy="download",
        )
    else:
        raise NotImplementedError(type(productionRequest))

    pr.outConfigName = "validation"
    pr.outputSEs = ["Tier1-Buffer"]

    if ancestorDepth is not None:
        kwargs["ancestorDepth"] = ancestorDepth

    kwargs["stepsInProd"] = _steps_to_production_dict(productionRequest.steps, inputFileType)
    kwargs["outputSE"] = {
        t["FileType"]: "Tier1-Buffer" for step in kwargs["stepsInProd"] for t in step["visibilityFlag"]
    }
    kwargs["priority"] = 0
    kwargs["cpu"] = 100

    return pr, kwargs


def _steps_to_production_dict(steps: list[ProductionStep], inputFileType: str | None) -> list[dict]:
    """Convert steps into list of dictionaries expected by ProductionRequest._buildProduction

    Normally this is handled by ProductionRequest.resolveSteps however this only
    supports reading from the bookkeeping.

    TODO: The ProductionRequest class should be refactored.
    """
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import step_to_step_manager_dict

    stepsInProd = []
    for i, dirac_step in enumerate(steps):
        result = step_to_step_manager_dict(i + 1, dirac_step)
        step_dict = result["Step"]
        step_dict["StepId"] = step_dict.get("StepId", 12345)
        if len(dirac_step.input) > 1 and inputFileType is None:
            raise NotImplementedError(
                f"Multiple input file types found, pick one of with --input-file-type:"
                f" {' '.join(repr(f.type) for f in dirac_step.input)}"
            )
        step_dict["fileTypesIn"] = [
            f.type for f in dirac_step.input if inputFileType is None or f.type == inputFileType
        ]
        if len(dirac_step.input) > 1:
            print(f"Assuming that step {i+1} is a merging step and reducing output filetypes to {inputFileType}")
            step_dict["fileTypesOut"] = [f.type for f in dirac_step.output if f.type == inputFileType]
            if len(step_dict["fileTypesOut"]) != 1:
                raise NotImplementedError(step_dict["fileTypesOut"])
        else:
            step_dict["fileTypesOut"] = [f.type for f in dirac_step.output]
        step_dict["ExtraPackages"] = ";".join([f"{d.name}.{d.version}" for d in dirac_step.data_pkgs])
        step_dict.setdefault("OptionsFormat", "")
        step_dict.setdefault("SystemConfig", "")
        step_dict.setdefault("mcTCK", "")
        step_dict["ExtraOptions"] = ""
        step_dict["visibilityFlag"] = result["OutputFileTypes"]
        step_dict["EventTimeout"] = None
        # Normally ProductionRequest.resolveSteps will set these but that only supports getting IDs from the bookkeeping
        for field in ["CONDDB", "DDDB", "DQTag"]:
            if step_dict[field] == "fromPreviousStep":
                step_dict[field] = stepsInProd[i - 1][field]
        stepsInProd.append(step_dict)
    return stepsInProd


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Agent.NagiosTopologyAgent.

NagiosTopologyAgent.__bases__: DIRAC.Core.Base.AgentModule.AgentModule

xml_append
"""
import os
import socket
import json

from urllib.parse import urlparse
from typing import Optional

from DIRAC import S_OK, gConfig, rootPath
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Resources.Storage.StorageElement import StorageElement

AGENT_NAME = "ResourceStatus/NagiosTopologyAgent"


MAPPING_CE_TYPE = {
    "lcg": "CE",
    "cream": "CREAM-CE",
    "arc": "ARC-CE",
    "arc6": "ARC-CE",
    "arex": "ARC-CE",
    "htcondorce": "HTCONDOR-CE",
    "vac": "VAC",
    "cloud": "CLOUD",
    "boinc": "BOINC",
    "vcycle": "VCYCLE",
    "dirac": "DIRAC",
}


def isHostIPV6(host: str) -> bool | None:
    """Test if the given host is ipv6 capable.

    :returns: True if IPv6 capable, False if only IPv4, None in case of resolution problem
    """
    try:  # First try IPV6
        socket.getaddrinfo(host, None, socket.AF_INET6)
        return True
    except socket.gaierror:  # No IPv6 address
        try:  # Next try IPv4
            socket.getaddrinfo(host, None, socket.AF_INET)
            return False
        except socket.gaierror:  # The host does not exist (no IPv6 or IPv4 address)
            return None


class NagiosTopologyAgent(AgentModule):
    """This agent loops over the Dirac CS and extracts the necessary information
    to create a "topology map" which is used by the IT provided Nagios system to
    test Grid sites. The topology information defines the services to be tested.

    NagiosTopologyAgent, writes the xml topology consumed by Nagios to run
    the tests.
    """

    def __init__(self, *args, **kwargs):
        AgentModule.__init__(self, *args, **kwargs)

        self.dryRun = False

    def execute(self):
        """Generates xml and json topology files"""

        res = getSites()
        if not res["OK"]:
            return res
        sites = res["Value"]

        # First, create the JSON topology, which contains also what the XML topology needs
        res = self.createJSONTopology(sites)
        if not res["OK"]:
            self.log.error("Failed to create JSON topology", res["Message"])
            return res
        self.log.info("Created JSON topology")

        return S_OK()

    def createJSONTopology(self, sites):
        """Function that creates a topology.json file, based on specification from WLCG group"""

        fullSitesDict = dict()
        for site in sites:
            grid = site.split(".")[0]
            res = gConfig.getOptionsDict(f"Resources/Sites/{grid}/{site}")
            if not res["OK"]:
                self.log.error("Failure getting options dict for site", f"{site}: {res['Message']}")
                continue
            siteInfoCS = res["Value"]
            wlcgName = siteInfoCS.get("Name")

            siteInfo = dict()
            siteInfo["WLCG_site"] = wlcgName
            # Tier level 4 is what they want for opportunistic resource
            siteInfo["Tier"] = siteInfoCS.get("MoUTierLevel", "4")
            siteInfo["Type"] = "GRID" if grid == "LCG" else site.split(".")[0]

            # Services is for CEs and SEs
            siteInfoServices = dict()

            # CEs
            res = gConfig.getSections(cfgPath("Resources", "Sites", grid, site, "CEs"), [])
            if not res["OK"]:
                self.log.error("Failure getting CEs section for site", f"{site}: {res['Message']}")
                continue
            if not res["Value"]:
                self.log.warn("Site without CEs", f"({site})")
            else:
                cesList = res["Value"]
                for ce in cesList:
                    res = gConfig.getOptionsDict(cfgPath("Resources", "Sites", grid, site, "CEs", ce))
                    if not res["OK"]:
                        self.log.error("Failure getting info on ce", f"{ce}: {res['Message']}")
                        continue

                    ceDetailsInCS = res["Value"]
                    ceDetails = dict()
                    ceDetails["type"] = "compute"
                    ceDetails["access_points"] = dict()
                    ceDetails["access_points"][ce] = dict()
                    ceDetails["access_points"][ce]["endpoint_url"] = ce
                    ceDetails["access_points"][ce]["ipv6_status"] = isHostIPV6(ce)
                    ceDetails["access_points"][ce]["type"] = MAPPING_CE_TYPE.get(
                        ceDetailsInCS["CEType"].lower(), "UNDEFINED"
                    )
                    # Port
                    if "Port" in ceDetailsInCS:
                        ceDetails["access_points"][ce]["port"] = ceDetailsInCS["Port"]
                    ceDetails["access_points"][ce]["monitored"] = "yes" if wlcgName else "no"
                    ceDetails["access_points"][ce]["quality_level"] = "production"
                    # queues
                    res = gConfig.getSections(cfgPath("Resources", "Sites", grid, site, "CEs", ce, "Queues"), [])
                    if not res["OK"]:
                        self.log.error("Failure getting info on ce/queues", f"{ce}: {res['Message']}")
                        continue
                    if not res["Value"]:
                        self.log.warn("CE without queues", f"({ce})")
                    else:
                        for queue in res["Value"]:
                            ceDetails["compute_shares"] = dict()
                            ceDetails["compute_shares"][queue] = dict()
                            ceDetails["compute_shares"][queue]["type"] = "queue"
                            ceDetails["compute_shares"][queue]["name"] = queue

                    siteInfoServices[ce] = ceDetails

            # SEs
            ses = gConfig.getValue(cfgPath("Resources", "Sites", grid, site, "SE"), [])
            if not ses:
                pass
            for se in ses:
                diracSE = StorageElement(se)  # this 'se' is .e.g. 'CERN-DST-EOS'
                seDetailsForDIRACSE = dict()
                seStorageSharesForDIRACSE = dict()
                writeProtocols = diracSE.localWriteProtocolList
                for diracSEoption in diracSE.protocolOptions:
                    # Building the storage_endpoints section
                    seDetails = dict()
                    proto = diracSEoption["Protocol"]
                    ep = os.path.join(
                        proto + "://" + diracSEoption["Host"].strip("/") + ":" + diracSEoption["Port"],
                        diracSEoption["Path"].strip("/"),
                    )
                    seDetails["endpoint_url"] = ep
                    seDetails["ipv6_status"] = isHostIPV6(urlparse(ep).hostname)
                    seDetails["interface_type"] = proto
                    seDetails["monitored"] = "yes" if wlcgName else "no"
                    seDetails["quality_level"] = "production"
                    seDetails["read_only"] = not (proto in writeProtocols)

                    dseep = "_".join([se, diracSEoption["Protocol"]])  # just a unique name e.g. 'CERN-DST-EOS_root'
                    seDetailsForDIRACSE[dseep] = seDetails

                    if diracSEoption.get("SpaceToken"):
                        # Building the storage_shares section
                        seDetails = dict()
                        seDetails["Name"] = diracSEoption["SpaceToken"]
                        seDetails["path"] = diracSEoption["WSUrl"]
                        if diracSEoption["Protocol"].lower() == "srm":
                            seDetails["assigned_endpoints"] = ep

                        dsess = "_".join(
                            [se, diracSEoption["SpaceToken"]]
                        )  # just a unique name e.g. 'CNAF-DST_LHCb-Disk'
                        seStorageSharesForDIRACSE[dsess] = seDetails

                siteInfoServices[se] = dict()
                siteInfoServices[se]["storage_endpoints"] = seDetailsForDIRACSE
                if seStorageSharesForDIRACSE:
                    siteInfoServices[se]["storage_shares"] = seStorageSharesForDIRACSE

                siteInfoServices[se]["storage_backend"] = {"T1D0": "tape", "T0D1": "disk"}.get(
                    diracSE.options.get("SEType"), "disk"
                )
                siteInfoServices[se]["type"] = "storage"
                siteInfoServices[se]["quality_level"] = "production"
                siteInfoServices[se]["monitored"] = "yes" if wlcgName else "no"

            siteInfo["Services"] = siteInfoServices
            fullSitesDict[site] = siteInfo

        topologyFile = os.path.join(rootPath, "webRoot/www/topology/topology.json")
        if not self.dryRun:
            with open(topologyFile, "w") as tj:
                json.dump(fullSitesDict, tj)
                self.log.verbose("JSON topology saved", topologyFile)
        else:
            self.log.info(f"[DRY RUN]Would have created file {topologyFile} with following content")
            self.log.info(json.dumps(fullSitesDict))

        return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test Interfaces API DiracProduction."""
import os

import LHCbDIRAC.Interfaces.API.DiracProduction as moduleTested
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob

lj = LHCbJob()


def test_LJ_setApplication():
    open("optionsFiles", "a").close()
    res = lj.setApplication("appName", "v1r0", "optionsFiles", systemConfig="x86_64-slc6-gcc-44-opt")
    assert res["OK"] is True
    res = lj.setApplication("appName", "v1r0", "optionsFiles", systemConfig="x86_64-slc5-gcc-41-opt")
    assert res["OK"] is True
    res = lj.setApplication("appName", "v1r0", "optionsFiles", systemConfig="x86_64-slc5-gcc-43-opt", eventTimeout=1000)
    assert res["OK"] is True
    os.remove("optionsFiles")


def test_instantiate():
    """tests that we can instantiate one object of the tested class."""
    testClass = moduleTested.DiracProduction
    prod = testClass(1)
    assert "DiracProduction" == prod.__class__.__name__



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import DIRAC
from DIRAC.Core.Base.Script import Script


def usage():
    """usage.

    Prints script usage
    """
    print(f"Usage: {Script.scriptName} <WMS Job ID> [<WMS Job ID>]")
    DIRAC.exit(2)


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    args = Script.getPositionalArgs()

    if len(args) < 1:
        usage()

    jobIDs = []
    diracProd = DiracProduction()
    try:
        jobIDs = [int(jobID) for jobID in args]
    except Exception as x:
        print("ERROR WMS JobID(s) must be integers")
        DIRAC.exit(2)

    exitCode = 0
    errorList = []

    for job in jobIDs:
        result = diracProd.getWMSProdJobID(job, printOutput=True)
        if "Message" in result:
            errorList.append((job, result["Message"]))
            exitCode = 2
        elif not result:
            errorList.append((job, "Null result for getWMSProdJobID() call"))
            exitCode = 2
        else:
            exitCode = 0

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()
